'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

''' Base classes for controls with items, such as ListCtrl, TreeCtrl, 
    and TreeListCtrl. ''' # pylint: disable=W0105


import wx, draganddrop, autowidth, tooltip, inspect
from taskcoachlib.thirdparty import hypertreelist


class _CtrlWithItemsMixin(object):
    ''' Base class for controls with items, such as ListCtrl, TreeCtrl,
        TreeListCtrl, etc. '''

    def _itemIsOk(self, item):
        try:
            return item.IsOk()          # for Tree(List)Ctrl
        except AttributeError:
            return item != wx.NOT_FOUND # for ListCtrl
        
    def _objectBelongingTo(self, item):
        if not self._itemIsOk(item):
            return None
        try:
            return self.GetItemPyData(item) # TreeListCtrl
        except AttributeError:
            return self.getItemWithIndex(item) # ListCtrl

    def SelectItem(self, item, *args, **kwargs):
        try:
            # Tree(List)Ctrl:
            super(_CtrlWithItemsMixin, self).SelectItem(item, *args, **kwargs)
        except AttributeError:
            # ListCtrl:
            select = kwargs.get('select', True)
            newState = wx.LIST_STATE_SELECTED
            if not select:
                newState = ~newState
            self.SetItemState(item, newState, wx.LIST_STATE_SELECTED)


class _CtrlWithPopupMenuMixin(_CtrlWithItemsMixin):
    ''' Base class for controls with popupmenu's. '''
    
    @staticmethod
    def _attachPopupMenu(eventSource, eventTypes, eventHandler):
        for eventType in eventTypes:
            eventSource.Bind(eventType, eventHandler)


class _CtrlWithItemPopupMenuMixin(_CtrlWithPopupMenuMixin):
    ''' Popupmenu's on items. '''

    def __init__(self, *args, **kwargs):
        self.__popupMenu = kwargs.pop('itemPopupMenu')
        super(_CtrlWithItemPopupMenuMixin, self).__init__(*args, **kwargs)
        if self.__popupMenu is not None:
            self._attachPopupMenu(self,
                (wx.EVT_TREE_ITEM_RIGHT_CLICK, wx.EVT_CONTEXT_MENU), 
                self.onItemPopupMenu)

    def onItemPopupMenu(self, event):
        # Make sure the window this control is in has focus:
        try:
            window = event.GetEventObject().MainWindow
        except AttributeError:
            window = event.GetEventObject()
        window.SetFocus()
        if hasattr(event, 'GetPoint'):
            # Make sure the item under the mouse is selected because that
            # is what users expect and what is most user-friendly. Not all
            # widgets do this by default, e.g. the TreeListCtrl does not.
            item = self.HitTest(event.GetPoint())[0]
            if not self._itemIsOk(item):
                return
            if not self.IsSelected(item):
                self.UnselectAll()
                self.SelectItem(item)
        self.PopupMenu(self.__popupMenu)


class _CtrlWithColumnPopupMenuMixin(_CtrlWithPopupMenuMixin):
    ''' This class enables a right-click popup menu on column headers. The 
        popup menu should expect a public property columnIndex to be set so 
        that the control can tell the menu which column the user clicked to
        popup the menu. '''
    
    def __init__(self, *args, **kwargs):
        self.__popupMenu = kwargs.pop('columnPopupMenu')
        super(_CtrlWithColumnPopupMenuMixin, self).__init__(*args, **kwargs)
        if self.__popupMenu is not None:
            self._attachPopupMenu(self, [wx.EVT_LIST_COL_RIGHT_CLICK],
                self.onColumnPopupMenu)
            
    def onColumnPopupMenu(self, event):
        # We store the columnIndex in the menu, because it's near to 
        # impossible for commands in the menu to determine on what column the
        # menu was popped up.
        columnIndex = event.GetColumn()
        self.__popupMenu.columnIndex = columnIndex
        # Because right-clicking on column headers does not automatically give
        # focus to the control, we force the focus:
        try:
            window = event.GetEventObject().GetMainWindow()
        except AttributeError:
            window = event.GetEventObject()
        window.SetFocus()
        self.PopupMenuXY(self.__popupMenu, *event.GetPosition())
        event.Skip(False)
        

class _CtrlWithDropTargetMixin(_CtrlWithItemsMixin):
    ''' Control that accepts files, e-mails or URLs being dropped onto items. '''
    
    def __init__(self, *args, **kwargs):
        self.__onDropURLCallback = kwargs.pop('onDropURL', None)
        self.__onDropFilesCallback = kwargs.pop('onDropFiles', None)
        self.__onDropMailCallback = kwargs.pop('onDropMail', None)
        super(_CtrlWithDropTargetMixin, self).__init__(*args, **kwargs)
        if self.__onDropURLCallback or self.__onDropFilesCallback or self.__onDropMailCallback:
            dropTarget = draganddrop.DropTarget(self.onDropURL,
                                                self.onDropFiles,
                                                self.onDropMail,
                                                self.onDragOver)
            self.GetMainWindow().SetDropTarget(dropTarget)

    def onDropURL(self, x, y, url):
        item = self.HitTest((x, y))[0]
        if self.__onDropURLCallback:
            self.__onDropURLCallback(self._objectBelongingTo(item), url)

    def onDropFiles(self, x, y, filenames):
        item = self.HitTest((x, y))[0]
        if self.__onDropFilesCallback:
            self.__onDropFilesCallback(self._objectBelongingTo(item), filenames)

    def onDropMail(self, x, y, mail):
        item = self.HitTest((x, y))[0]
        if self.__onDropMailCallback:
            self.__onDropMailCallback(self._objectBelongingTo(item), mail)

    def onDragOver(self, x, y, defaultResult):
        item, flags = self.HitTest((x, y))[:2]
        if self._itemIsOk(item):
            if flags & wx.TREE_HITTEST_ONITEMBUTTON:
                self.Expand(item)
        return defaultResult

    def GetMainWindow(self):
        try:
            return super(_CtrlWithDropTargetMixin, self).GetMainWindow()
        except AttributeError:
            return self


class CtrlWithToolTipMixin(_CtrlWithItemsMixin, tooltip.ToolTipMixin):
    ''' Control that has a different tooltip for each item '''
    def __init__(self, *args, **kwargs):
        super(CtrlWithToolTipMixin, self).__init__(*args, **kwargs)
        self.__tip = tooltip.SimpleToolTip(self)

    def OnBeforeShowToolTip(self, x, y): 
        item, _, column = self.HitTest(wx.Point(x, y))
        domainObject = self._objectBelongingTo(item)
        if domainObject:
            tooltipData = self.getItemTooltipData(domainObject)
            doShow = any([data[1] for data in tooltipData])
            if doShow:
                self.__tip.SetData(tooltipData)
                return self.__tip
        return None


class CtrlWithItemsMixin(_CtrlWithItemPopupMenuMixin, _CtrlWithDropTargetMixin):
    pass


class Column(object):
    def __init__(self, name, columnHeader, *eventTypes, **kwargs):
        self.__name = name
        self.__columnHeader = columnHeader
        self.width = kwargs.pop('width', hypertreelist._DEFAULT_COL_WIDTH) # pylint: disable=W0212
        # The event types to use for registering an observer that is
        # interested in changes that affect this column:
        self.__eventTypes = eventTypes
        self.__sortCallback = kwargs.pop('sortCallback', None)
        self.__renderCallback = kwargs.pop('renderCallback',
            self.defaultRenderer)
        self.__resizeCallback = kwargs.pop('resizeCallback', None)
        self.__alignment = kwargs.pop('alignment', wx.LIST_FORMAT_LEFT)
        self.__hasImages = 'imageIndicesCallback' in kwargs
        self.__imageIndicesCallback = kwargs.pop('imageIndicesCallback',
            self.defaultImageIndices) or self.defaultImageIndices
        # NB: because the header image is needed for sorting a fixed header
        # image cannot be combined with a sortable column
        self.__headerImageIndex = kwargs.pop('headerImageIndex', -1)
        self.__editCallback = kwargs.get('editCallback', None)
        self.__editControlClass = kwargs.get('editControl', None)
        self.__parse = kwargs.get('parse', lambda value: value)
        self.__settings = kwargs.get('settings', None) # FIXME: Column shouldn't need to know about settings
        
    def name(self):
        return self.__name
        
    def header(self):
        return self.__columnHeader
    
    def headerImageIndex(self):
        return self.__headerImageIndex

    def eventTypes(self):
        return self.__eventTypes
    
    def setWidth(self, width):
        self.width = width
        if self.__resizeCallback:
            self.__resizeCallback(self, width)

    def sort(self, *args, **kwargs):
        if self.__sortCallback:
            self.__sortCallback(*args, **kwargs)

    def __filterArgs(self, func, kwargs):
        actualKwargs = dict()
        argNames = inspect.getargspec(func).args
        return dict([(name, value) for name, value in kwargs.items() if name in argNames])

    def render(self, *args, **kwargs):
        return self.__renderCallback(*args, **self.__filterArgs(self.__renderCallback, kwargs))

    def defaultRenderer(self, *args, **kwargs): # pylint: disable=W0613
        return unicode(args[0])

    def alignment(self):
        return self.__alignment
    
    def defaultImageIndices(self, *args, **kwargs): # pylint: disable=W0613
        return {wx.TreeItemIcon_Normal: -1}
        
    def imageIndices(self, *args, **kwargs):
        return self.__imageIndicesCallback(*args, **kwargs)
    
    def hasImages(self):
        return self.__hasImages
    
    def isEditable(self):
        return self.__editControlClass != None and self.__editCallback != None
    
    def onEndEdit(self, item, newValue):
        self.__editCallback(item, newValue)
                
    def editControl(self, parent, item, columnIndex, domainObject):
        value = self.value(domainObject)
        kwargs = dict(settings=self.__settings) if self.__settings else dict()
        # pylint: disable=W0142
        return self.__editControlClass(parent, wx.ID_ANY, item, columnIndex,
                                       parent, value, **kwargs)
    
    def parse(self, value):
        return self.__parse(value)
    
    def value(self, domainObject):
        return getattr(domainObject, self.name())()
    
    def __eq__(self, other):
        return self.name() == other.name()
        

class _BaseCtrlWithColumnsMixin(object):
    ''' A base class for all controls with columns. Note that this class and 
        its subclasses do not support addition or deletion of columns after 
        the initial setting of columns. '''

    def __init__(self, *args, **kwargs):
        self.__allColumns = kwargs.pop('columns')
        super(_BaseCtrlWithColumnsMixin, self).__init__(*args, **kwargs)
        # This  is  used to  keep  track  of  which column  has  which
        # index. The only  other way would be (and  was) find a column
        # using its header, which causes problems when several columns
        # have the same header. It's a list of (index, column) tuples.
        self.__indexMap = []
        self._setColumns()

    def _setColumns(self):
        for columnIndex, column in enumerate(self.__allColumns):
            self._insertColumn(columnIndex, column)
            
    def _insertColumn(self, columnIndex, column):
        newMap = []
        for colIndex, col in self.__indexMap:
            if colIndex >= columnIndex:
                newMap.append((colIndex + 1, col))
            else:
                newMap.append((colIndex, col))
        newMap.append((columnIndex, column))
        self.__indexMap = newMap

        self.InsertColumn(columnIndex, column.header() if column.headerImageIndex() == -1 else u'', 
            format=column.alignment(), width=column.width)

        columnInfo = self.GetColumn(columnIndex)
        columnInfo.SetImage(column.headerImageIndex())
        self.SetColumn(columnIndex, columnInfo)

    def _deleteColumn(self, columnIndex):
        newMap = []
        for colIndex, col in self.__indexMap:
            if colIndex > columnIndex:
                newMap.append((colIndex - 1, col))
            elif colIndex < columnIndex:
                newMap.append((colIndex, col))
        self.__indexMap = newMap
        self.DeleteColumn(columnIndex)

    def _allColumns(self):
        return self.__allColumns

    def _getColumn(self, columnIndex):
        for colIndex, col in self.__indexMap:
            if colIndex == columnIndex:
                return col
        raise IndexError
   
    def _getColumnHeader(self, columnIndex):
        ''' The currently displayed column header in the column with index 
            columnIndex. '''
        return self.GetColumn(columnIndex).GetText()

    def _getColumnIndex(self, column):
        ''' The current column index of the column 'column'. '''
        try:
            return self.__allColumns.index(column) # Uses overriden __eq__
        except ValueError:
            raise ValueError, '%s: unknown column' % column.name()

        
class _CtrlWithHideableColumnsMixin(_BaseCtrlWithColumnsMixin):        
    ''' This class supports hiding columns. '''
    
    def showColumn(self, column, show=True):
        ''' showColumn shows or hides the column for column. 
            The column is actually removed or inserted into the control because 
            although TreeListCtrl supports hiding columns, ListCtrl does not. 
            '''
        columnIndex = self._getColumnIndex(column)
        if show and not self.isColumnVisible(column):
            self._insertColumn(columnIndex, column)
        elif not show and self.isColumnVisible(column):
            self._deleteColumn(columnIndex)

    def isColumnVisible(self, column):
        return column in self._visibleColumns()

    def _getColumnIndex(self, column):
        ''' _getColumnIndex returns the actual columnIndex of the column if it 
            is visible, or the position it would have if it were visible. '''
        columnIndexWhenAllColumnsVisible = super(_CtrlWithHideableColumnsMixin, self)._getColumnIndex(column)
        for columnIndex, visibleColumn in enumerate(self._visibleColumns()):
            if super(_CtrlWithHideableColumnsMixin, self)._getColumnIndex(visibleColumn) >= columnIndexWhenAllColumnsVisible:
                return columnIndex
        return self.GetColumnCount() # Column header not found

    def _visibleColumns(self):
        return [self._getColumn(columnIndex) for columnIndex in range(self.GetColumnCount())]


class _CtrlWithSortableColumnsMixin(_BaseCtrlWithColumnsMixin):
    ''' This class adds sort indicators and clickable column headers that 
        trigger callbacks to (re)sort the contents of the control. '''
    
    def __init__(self, *args, **kwargs):
        super(_CtrlWithSortableColumnsMixin, self).__init__(*args, **kwargs)
        self.Bind(wx.EVT_LIST_COL_CLICK, self.onColumnClick)
        self.__currentSortColumn = self._getColumn(0)
        self.__currentSortImageIndex = -1
                
    def onColumnClick(self, event):
        event.Skip(False)
        # Make sure the window this control is in has focus:
        try:
            window = event.GetEventObject().GetMainWindow()
        except AttributeError:
            window = event.GetEventObject()
        window.SetFocus()
        columnIndex = event.GetColumn()
        if 0 <= columnIndex < self.GetColumnCount():
            column = self._getColumn(columnIndex)
            # Use CallAfter to make sure the window this control is in is
            # activated before we process the column click:
            wx.CallAfter(column.sort, event)
        
    def showSortColumn(self, column):
        if column != self.__currentSortColumn:
            self._clearSortImage()
        self.__currentSortColumn = column
        self._showSortImage()

    def showSortOrder(self, imageIndex):
        self.__currentSortImageIndex = imageIndex
        self._showSortImage()
                
    def _clearSortImage(self):
        self.__setSortColumnImage(-1)
    
    def _showSortImage(self):
        self.__setSortColumnImage(self.__currentSortImageIndex)
            
    def _currentSortColumn(self):
        return self.__currentSortColumn
        
    def __setSortColumnImage(self, imageIndex):
        columnIndex = self._getColumnIndex(self.__currentSortColumn)
        columnInfo = self.GetColumn(columnIndex)
        if columnInfo.GetImage() == imageIndex:
            pass # The column is already showing the right image, so we're done
        else:
            columnInfo.SetImage(imageIndex)
            self.SetColumn(columnIndex, columnInfo)


class _CtrlWithAutoResizedColumnsMixin(autowidth.AutoColumnWidthMixin):
    def __init__(self, *args, **kwargs):
        super(_CtrlWithAutoResizedColumnsMixin, self).__init__(*args, **kwargs)
        self.Bind(wx.EVT_LIST_COL_END_DRAG, self.onEndColumnResize)
        
    def onEndColumnResize(self, event):
        ''' Save the column widths after the user did a resize. '''
        for index, column in enumerate(self._visibleColumns()):
            column.setWidth(self.GetColumnWidth(index))
        event.Skip()
        

class CtrlWithColumnsMixin(_CtrlWithAutoResizedColumnsMixin, 
                           _CtrlWithHideableColumnsMixin,
                           _CtrlWithSortableColumnsMixin, 
                           _CtrlWithColumnPopupMenuMixin):
    ''' CtrlWithColumnsMixin combines the functionality of its four parent 
        classes: automatic resizing of columns, hideable columns, columns with 
        sort indicators, and column popup menu's. '''
        
    def showColumn(self, column, show=True):
        super(CtrlWithColumnsMixin, self).showColumn(column, show)
        # Show sort indicator if the column that was just made visible is being sorted on
        if show and column == self._currentSortColumn():
            self._showSortImage()
            
    def _clearSortImage(self):
        # Only clear the sort image if the column in question is visible
        if self.isColumnVisible(self._currentSortColumn()):
            super(CtrlWithColumnsMixin, self)._clearSortImage()
            
    def _showSortImage(self):
        # Only show the sort image if the column in question is visible
        if self.isColumnVisible(self._currentSortColumn()):
            super(CtrlWithColumnsMixin, self)._showSortImage()
            




# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

''' render.py - functions to render various objects, like date, time, 
etc. '''  # pylint: disable=W0105

from taskcoachlib.domain import date as datemodule
from taskcoachlib.thirdparty import desktop
from taskcoachlib.i18n import _
from taskcoachlib import operating_system
import datetime
import codecs
import locale
import re

# pylint: disable=W0621


def priority(priority):
    ''' Render an (integer) priority '''
    return str(priority)

 
def timeLeft(time_left, completed_task):
    ''' Render time left as a text string. Returns an empty string for 
        completed tasks and for tasks without planned due date. Otherwise it 
        returns the number of days, hours, and minutes left. '''
    if completed_task or time_left == datemodule.TimeDelta.max:
        return ''
    sign = '-' if time_left.days < 0 else ''
    time_left = abs(time_left)
    if time_left.days > 0:
        days = _('%d days') % time_left.days if time_left.days > 1 else \
               _('1 day')
        days += ', '
    else:
        days = '' 
    hours_and_minutes = ':'.join(str(time_left).split(':')[:-1]).split(', ')[-1]
    return sign + days + hours_and_minutes


def timeSpent(timeSpent, showSeconds=True, decimal=False):
    ''' Render time spent (of type date.TimeDelta) as
        "<hours>:<minutes>:<seconds>" or "<hours>:<minutes>" '''
    if decimal:
        return timeSpentDecimal(timeSpent)

    zero = datemodule.TimeDelta()
    if timeSpent == zero:
        return ''
    else:
        sign = '-' if timeSpent < zero else ''
        hours, minutes, seconds = timeSpent.hoursMinutesSeconds()
        return sign + '%d:%02d' % (hours, minutes) + \
               (':%02d' % seconds if showSeconds else '')

def timeSpentDecimal(timeSpent):
    ''' Render time spent (of type date.TimeDelta) as
        "<hours>.<fractional hours> '''
    zero = datemodule.TimeDelta()
    if timeSpent == zero:
        return ''
    else:
        sign = '-' if timeSpent < zero else ''
        hours, minutes, seconds = timeSpent.hoursMinutesSeconds()
        decimalHours = hours + minutes/60.0 + seconds/3600.0
        return sign + '%.2f' % (decimalHours)

def recurrence(recurrence):
    ''' Render the recurrence as a short string describing the frequency of
        the recurrence. '''
    if not recurrence:
        return ''
    if recurrence.amount > 2:
        labels = [_('Every %(frequency)d days'), 
                  _('Every %(frequency)d weeks'),
                  _('Every %(frequency)d months'),
                  _('Every %(frequency)d years')] 
    elif recurrence.amount == 2:
        labels = [_('Every other day'), _('Every other week'),
                  _('Every other month'), _('Every other year')]
    else:
        labels = [_('Daily'), _('Weekly'), _('Monthly'), _('Yearly')] 
    mapping = dict(zip(['daily', 'weekly', 'monthly', 'yearly'], labels))
    return mapping.get(recurrence.unit) % dict(frequency=recurrence.amount)


def budget(aBudget):
    ''' Render budget (of type date.TimeDelta) as 
        "<hours>:<minutes>:<seconds>". '''
    return timeSpent(aBudget)


# Default time formatting
language_and_country = locale.getlocale(locale.LC_TIME)[0]
if language_and_country and ('_US' in language_and_country or 
                             '_United States' in language_and_country):
    timeFormat = '%I %p'
    timeWithMinutesFormat = '%I:%M %p'
    timeWithSecondsFormat = '%I:%M:%S %p'
else:
    timeFormat = '%H'
    timeWithMinutesFormat = '%H:%M'  # %X includes seconds (see http://stackoverflow.com/questions/2507726)
    timeWithSecondsFormat = '%X'
def rawTimeFunc(dt, minutes=True, seconds=False):
    if seconds:
        fmt = timeWithSecondsFormat
    else:
        if minutes:
            fmt = timeWithMinutesFormat
        else:
            fmt = timeFormat
    return datemodule.DateTime.strftime(dt, fmt)


dateFormat = '%x'
def rawDateFunc(dt=None):
    return operating_system.decodeSystemString(datetime.datetime.strftime(dt, dateFormat))


def dateFunc(dt=None, humanReadable=False):
    if humanReadable:
        theDate = dt.date()
        if theDate == datemodule.Now().date():
            return _('Today')
        elif theDate == datemodule.Yesterday().date():
            return _('Yesterday')
        elif theDate == datemodule.Tomorrow().date():
            return _('Tomorrow')
    return rawDateFunc(dt)


# OS-specific time formatting
if operating_system.isWindows():
    import pywintypes, win32api
    def rawTimeFunc(dt, minutes=True, seconds=False):
        if seconds:
            # You can't include seconds without minutes
            flags = 0x0
        else:
            if minutes:
                flags = 0x2
            else:
                flags = 0x1
        return operating_system.decodeSystemString(win32api.GetTimeFormat(0x400, flags, None if dt is None else pywintypes.Time(dt), None))

    def rawDateFunc(dt):
        return operating_system.decodeSystemString(win32api.GetDateFormat(0x400, 0, None if dt is None else pywintypes.Time(dt), None))
elif operating_system.isMac():
    import Cocoa, calendar
    # We don't actually respect the 'seconds' parameter; this assumes that the short time format does
    # not include them, but the medium format does.
    _shortFormatter = Cocoa.NSDateFormatter.alloc().init()
    _shortFormatter.setFormatterBehavior_(Cocoa.NSDateFormatterBehavior10_4)
    _shortFormatter.setTimeStyle_(Cocoa.NSDateFormatterShortStyle)
    _shortFormatter.setDateStyle_(Cocoa.NSDateFormatterNoStyle)
    _shortFormatter.setTimeZone_(Cocoa.NSTimeZone.timeZoneForSecondsFromGMT_(0))
    _mediumFormatter = Cocoa.NSDateFormatter.alloc().init()
    _mediumFormatter.setFormatterBehavior_(Cocoa.NSDateFormatterBehavior10_4)
    _mediumFormatter.setTimeStyle_(Cocoa.NSDateFormatterMediumStyle)
    _mediumFormatter.setDateStyle_(Cocoa.NSDateFormatterNoStyle)
    _mediumFormatter.setTimeZone_(Cocoa.NSTimeZone.timeZoneForSecondsFromGMT_(0))
    # Special case for hour without minutes or seconds. I don't know if it is possible to get the AM/PM
    # setting alone, so parse the format string instead.
    # See http://www.unicode.org/reports/tr35/tr35-25.html#Date_Format_Patterns
    _state = 0
    _hourFormat = u''
    _ampmFormat = u''
    for c in _mediumFormatter.dateFormat():
        if _state == 0:
            if c == u"'":
                _state = 1 # After single quote
            elif c in [u'h', u'H', u'k', u'K', u'j']:
                _hourFormat += c
            elif c == 'a':
                _ampmFormat = c
        elif _state == 1:
            if c == u"'":
                _state = 0
            else:
                _state = 2 # Escaped string
        elif _state == 2:
            if c == u"'":
                _state = 0
    _hourFormatter = Cocoa.NSDateFormatter.alloc().init()
    _hourFormatter.setFormatterBehavior_(Cocoa.NSDateFormatterBehavior10_4)
    _hourFormatter.setDateFormat_(_hourFormat + (' %s' % _ampmFormat if _ampmFormat else ''))
    _hourFormatter.setTimeZone_(Cocoa.NSTimeZone.timeZoneForSecondsFromGMT_(0))
    _dateFormatter = Cocoa.NSDateFormatter.alloc().init()
    _dateFormatter.setFormatterBehavior_(Cocoa.NSDateFormatterBehavior10_4)
    _dateFormatter.setDateStyle_(Cocoa.NSDateFormatterShortStyle)
    _dateFormatter.setTimeStyle_(Cocoa.NSDateFormatterNoStyle)
    _dateFormatter.setTimeZone_(Cocoa.NSTimeZone.timeZoneForSecondsFromGMT_(0))

    def _applyFormatter(dt, fmt):
        dt_native = Cocoa.NSDate.dateWithTimeIntervalSince1970_((dt - datetime.datetime(1970, 1, 1, 0, 0, 0, 0)).total_seconds())
        return fmt.stringFromDate_(dt_native)

    def rawTimeFunc(dt, minutes=True, seconds=False):
        if minutes:
            if seconds:
                return _applyFormatter(dt, _mediumFormatter)
            return _applyFormatter(dt, _shortFormatter)
        return _applyFormatter(dt, _hourFormatter)

    def rawDateFunc(dt):
        return _applyFormatter(datetime.datetime.combine(dt, datetime.time(0, 0, 0, 0)), _dateFormatter)
elif desktop.get_desktop() == 'KDE4':
    try:
        # Import gtk first because when it's imported indirectly it generates a RuntimeWarning.
        import gtk
        from PyKDE4.kdecore import KGlobal, KLocale
        from PyQt4.QtCore import QTime, QDate
    except ImportError:
        pass
    else:
        _localeCopy = KLocale(KGlobal.locale())
        if '%p' in KGlobal.locale().timeFormat():
            _localeCopy.setTimeFormat('%I %p')
        else:
            _localeCopy.setTimeFormat('%H')
        def rawTimeFunc(dt, minutes=True, seconds=False):
            qtdt = QTime(dt.hour, dt.minute, dt.second)
            if minutes:
                return unicode(KGlobal.locale().formatTime(qtdt, seconds))
            return unicode(_localeCopy.formatTime(qtdt))

        def rawDateFunc(dt):
            qtdt = QDate(dt.year, dt.month, dt.day)
            return unicode(KGlobal.locale().formatDate(qtdt, 0))


timeFunc = lambda dt, minutes=True, seconds=False: operating_system.decodeSystemString(rawTimeFunc(dt, minutes=minutes, seconds=seconds))

dateTimeFunc = lambda dt=None, humanReadable=False: u'%s %s' % (dateFunc(dt, humanReadable=humanReadable), timeFunc(dt))


def date(aDateTime, humanReadable=False):
    ''' Render a date/time as date. '''
    if str(aDateTime) == '':
        return ''
    year = aDateTime.year
    if year >= 1900:
        return dateFunc(aDateTime, humanReadable=humanReadable)
    else:
        result = date(datemodule.DateTime(year + 1900, aDateTime.month, 
                                          aDateTime.day), 
                      humanReadable=humanReadable)
        return re.sub(str(year + 1900), str(year), result)


def dateTime(aDateTime, humanReadable=False):
    if not aDateTime or aDateTime == datemodule.DateTime() or aDateTime == datemodule.DateTime.min:
        return ''
    timeIsMidnight = (aDateTime.hour, aDateTime.minute) in ((0, 0), (23, 59))
    year = aDateTime.year
    if year >= 1900:
        return dateFunc(aDateTime, humanReadable=humanReadable) if timeIsMidnight else \
            dateTimeFunc(aDateTime, humanReadable=humanReadable)
    else:
        result = dateTime(aDateTime.replace(year=year + 1900), humanReadable=humanReadable)
        return re.sub(str(year + 1900), str(year), result)

   
def dateTimePeriod(start, stop, humanReadable=False):
    if stop is None:
        return '%s - %s' % (dateTime(start, humanReadable=humanReadable), _('now'))
    elif start.date() == stop.date():
        return '%s %s - %s' % (date(start, humanReadable=humanReadable), 
                               time(start), time(stop))
    else:
        return '%s - %s' % (dateTime(start, humanReadable=humanReadable), dateTime(stop, humanReadable=humanReadable))
    
    
def time(dateTime, seconds=False, minutes=True):
    try:
        # strftime doesn't handle years before 1900, be prepared:
        dateTime = dateTime.replace(year=2000)  
    except TypeError:  # We got a time instead of a dateTime
        dateTime = datemodule.Now().replace(hour=dateTime.hour, 
                                            minute=dateTime.minute,
                                            second=dateTime.second) 
    return timeFunc(dateTime, minutes=minutes, seconds=seconds)


def month(dateTime):
    return dateTime.strftime('%Y %B')

    
def weekNumber(dateTime):
    # Would have liked to use dateTime.strftime('%Y-%U'), but the week number 
    # is one off in 2004
    return '%d-%d' % (dateTime.year, dateTime.weeknumber())


def monetaryAmount(aFloat):
    ''' Render a monetary amount, using the user's locale. '''
    return '' if round(aFloat, 2) == 0 else \
        locale.format('%.2f', aFloat, monetary=True)


def percentage(aFloat):
    ''' Render a percentage. '''
    return '' if round(aFloat, 0) == 0 else '%.0f%%' % aFloat


def exception(exception, instance):
    ''' Safely render an exception, being prepared for new exceptions. '''

    try:
        # In this order. Python 2.6 fixed the unicode exception problem.
        try:
            return unicode(instance)
        except UnicodeDecodeError:
            # On Windows, some exceptions raised by win32all lead to this
            # Hack around it
            result = []
            for val in instance.args:
                if isinstance(val, unicode):
                    result.append(val.encode('UTF-8'))
                else:
                    result.append(val)
            return unicode(result)
    except UnicodeEncodeError:
        return '<class %s>' % str(exception)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, wx
from taskcoachlib.gui import splash, icons


class SplashTest(test.wxTestCase):
    def setUp(self):
        super(SplashTest, self).setUp()
        self.splashScreen = splash.SplashScreen()
        self.splashScreen.Hide()

    def tearDown(self):
        self.splashScreen.Destroy()
        super(SplashTest, self).tearDown()

    def assertCorrectBitmap(self):
        expectedBitmap = icons.catalog['splash'].getBitmap()
        actualBitmap = self.splashScreen.GetSplashWindow().GetBitmap()
        bitmapData = lambda bitmap: bitmap.ConvertToImage().GetData()
        self.assertEqual(bitmapData(expectedBitmap), bitmapData(actualBitmap))

    def testTimeout(self):
        self.assertEqual(4000, self.splashScreen.GetTimeout())

    def testBitmap(self):
        self.assertCorrectBitmap()

    def testMirrorBitmapBackWhenLanguageIsRTL(self):
        class FakeModule(object):
            def currentLanguageIsRightToLeft(self):
                return True
        splash.i18n = FakeModule()
        self.assertCorrectBitmap()


'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain.note import Note
from taskcoachlib.domain.category import Category
from taskcoachlib.syncml.basesource import BaseSource
from taskcoachlib.persistence.icalendar import ical

from taskcoachlib.i18n import _

import wx, inspect

class NoteSource(BaseSource):
    def __init__(self, callback, noteList, categoryList, dataType, *args, **kwargs):
        super(NoteSource, self).__init__(callback, noteList, *args, **kwargs)

        self.categoryList = categoryList
        self._dataType = dataType

    def updateItemProperties(self, item, note):
        item.dataType = self._dataType
        if self._dataType == 'text/x-vnote':
            item.data = ical.VNoteFromNote(note, doFold=False).encode('UTF-8')
        else:
            item.data = (u'%s\n%s' % (note.subject(), note.description())).encode('UTF-8')

    def _parseObject(self, item):
        # Horde doesn't seem to give a fuck about the supported types we send it.
        if item.dataType == 'text/plain':
            lines = map(lambda x: x.rstrip('\r'), item.data.split('\n'))
            kwargs = dict(subject=lines[0], description='\n'.join(lines[1:])) if lines else dict()
            categories = list()
        else:
            parser = ical.VCalendarParser()
            parser.parse(map(lambda x: x.rstrip('\r'), item.data.strip().split('\n')))
            categories = parser.notes[0].pop('categories', [])

            kwargs = dict([(k, v) for k, v in parser.notes[0].items() if k in ['subject', 'description', 'id']])

        note = Note(**kwargs)
        for category in categories:
            categoryObject = self.categoryList.findCategoryByName(category)
            if categoryObject is None:
                categoryObject = Category(category)
                self.categoryList.extend([categoryObject])
            note.addCategory(categoryObject)
        return note

    def doUpdateItem(self, note, local):
        local.setSubject(note.subject())
        local.setDescription(note.description())

        for category in local.categories():
            category.removeCategorizable(local)

        local.setCategories(note.categories())

        for category in local.categories():
            category.addCategorizable(local)

        super(NoteSource, self).doUpdateItem(note, local)

        return 200



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.thirdparty.pubsub import pub
import wx


class AutoSaver(object):
    ''' AutoSaver observes task files. If a task file is changed by the user 
        (gets 'dirty') and auto save is on, AutoSaver saves the task file. '''
        
    def __init__(self, settings, *args, **kwargs):
        super(AutoSaver, self).__init__(*args, **kwargs)
        self.__settings = settings
        self.__task_files = set()
        self.__bound = False
        pub.subscribe(self.onTaskFileDirty, 'taskfile.dirty')

    def onTaskFileDirty(self, taskFile):
        ''' When a task file gets dirty and auto save is on, note it so 
            it can be saved during idle time. '''
        if self._needSave(taskFile):
            self.__task_files.add(taskFile)
        if not self.__bound:
            self.__bound = True
            wx.GetApp().Bind(wx.EVT_IDLE, self.on_idle)

    def _needSave(self, task_file):
        ''' Return whether the task file needs to be saved. '''
        return task_file.filename() and task_file.needSave() and \
            self.__settings.getboolean('file', 'autosave')

    def _needLoad(self, taskFile):
        return taskFile.changedOnDisk() and \
            self.__settings.getboolean('file', 'autoload')

    def on_idle(self, event):
        ''' Actually save the dirty files during idle time. '''
        event.Skip()
        wx.GetApp().Unbind(wx.EVT_IDLE, handler=self.on_idle)
        self.__bound = False
        while self.__task_files:
            task_file = self.__task_files.pop()
            if self._needSave(task_file):
                task_file.save()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, datetime
from taskcoachlib.domain import date


class PyDateTimeTest(test.TestCase):
    def testReplaceCannotBeEasilyUsedToFindTheLastDayofTheMonth(self):
        testDate = datetime.date(2004, 4, 1) # April 1st, 2004
        try:
            lastDayOfApril = testDate.replace(day=31)
            self.fail('Surprise! datetime.date.replace works as we want!') # pragma: no cover
            self.assertEqual(datetime.date(2004, 4, 30), lastDayOfApril) # pragma: no cover
        except ValueError:
            pass


class DateTimeTest(test.TestCase):
    def testWeekNumber(self):
        self.assertEqual(53, date.DateTime(2005,1,1).weeknumber())
        self.assertEqual(1, date.DateTime(2005,1,3).weeknumber())   
        
    def testStartOfDay(self):
        startOfDay = date.DateTime(2005,1,1,0,0,0,0)
        noonish = date.DateTime(2005,1,1,12,30,15,400)
        self.assertEqual(startOfDay, noonish.startOfDay())
        
    def testEndOfDay(self):
        endOfDay = date.DateTime(2005,1,1,23,59,59,999999)
        noonish = date.DateTime(2005,1,1,12,30,15,400)
        self.assertEqual(endOfDay, noonish.endOfDay())
        
    def testStartOfWorkWeekOnWednesday(self):
        startOfWorkWeek = date.DateTime(2011,7,25,0,0,0,0)
        wednesday = date.DateTime(2011,7,27,8,39,10)
        self.assertEqual(startOfWorkWeek, wednesday.startOfWorkWeek())
        
    def testStartOfWorkWeekOnMonday(self):
        startOfWorkWeek = date.DateTime(2011,7,25,0,0,0,0)
        monday = date.DateTime(2011,7,25,8,39,10)
        self.assertEqual(startOfWorkWeek, monday.startOfWorkWeek())

    def testStartOfWorkWeekOnSunday(self):
        startOfWorkWeek = date.DateTime(2011,7,18,0,0,0,0)
        sunday = date.DateTime(2011,7,24,8,39,10)
        self.assertEqual(startOfWorkWeek, sunday.startOfWorkWeek())
        
    def testEndOfWorkWeek(self):
        endOfWorkWeek = date.DateTime(2010,5,7,23,59,59,999999)
        midweek = date.DateTime(2010,5,5,12,30,15,200000)
        self.assertEqual(endOfWorkWeek, midweek.endOfWorkWeek())

    def testEndOfWorkWeek_OnSaturday(self):
        endOfWorkWeek = date.DateTime(2010,5,7,23,59,59,999999)
        midweek = date.DateTime(2010,5,1,12,30,15,200000)
        self.assertEqual(endOfWorkWeek, midweek.endOfWorkWeek())
        
    def testLastDayOfCurrentMonth_InFebruary2004(self):
        expected = date.DateTime(2004, 2, 29)
        actual = date.LastDayOfCurrentMonth(localtime=lambda: (2004, 2, 1))
        self.assertEqual(expected, actual)

    def testLastDayOfCurrentMonth_InDecember(self):
        expected = date.DateTime(2003, 12, 31)
        actual = date.LastDayOfCurrentMonth(localtime=lambda: (2003, 12, 1))
        self.assertEqual(expected, actual)

    def testFormat1900(self):
        self.assertEqual(date.DateTime(2, 5, 19, 0, 0, 0).strftime('%Y%m%d'), '20519')



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from wx.lib import hyperlink
from taskcoachlib import meta
from taskcoachlib.i18n import _
from wx.lib import sized_controls


class MessageDialog(sized_controls.SizedDialog):  # pylint: disable=R0904,R0901
    ''' Dialog for showing messages from the developers. '''
    def __init__(self, *args, **kwargs):
        self.__settings = kwargs.pop('settings')
        self.__message = kwargs.pop('message')
        self.__url = kwargs.pop('url')
        super(MessageDialog, self).__init__(title=_('Message from the %s '
                                            'developers') % meta.data.name, 
                                            *args, **kwargs)
        pane = self.GetContentsPane()
        pane.SetSizerType("vertical")
        self.__create_message(pane)
        button_sizer = self.CreateStdDialogButtonSizer(wx.OK)
        self.SetButtonSizer(button_sizer)
        self.Fit()
        self.CentreOnParent()
        button_sizer.GetAffirmativeButton().Bind(wx.EVT_BUTTON, self.on_close)
        self.Bind(wx.EVT_CLOSE, self.on_close)
        
    def __create_message(self, pane):
        ''' Create the interior parts of the dialog, i.e. the message for the
            user. '''
        message = wx.StaticText(pane, label=self.__message)
        message.Wrap(500)
        url_panel = sized_controls.SizedPanel(pane)
        url_panel.SetSizerType('horizontal')
        wx.StaticText(url_panel, label=_('See:'))
        hyperlink.HyperLinkCtrl(url_panel, label=self.__url)
        
    def current_message(self):
        ''' Return the currently displayed message. '''
        return self.__message
    
    def current_url(self):
        ''' Return the currently displayed url. '''
        return self.__url
        
    def on_close(self, event):
        ''' When the user closes the dialog, remember what the last displayed 
            message was. '''
        event.Skip()
        self.__settings.settext('view', 'lastdevelopermessage', self.__message)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os
from distutils import errors
import bdist_portable_base


class bdist_portableapps(bdist_portable_base.bdist_portable_base):
    
    description = 'create a PortableApps package'
    
    user_options = [
        ('bdist-base=', None, 
         'base directory for creating built distributions [build]'),
        ('dist-dir=', 'd', 'directory to put final package files in [dist]'),
        ('name=', None, 'name of the application'),
        ('version=', None, 'version of the application'),
        ('license=', None, 'license (title) of the application'),
        ('url=', None, 'url of the application homepage'),
        ('filename=', None, 'filename of the application without extension'),
        ('date=', None, 'the release date')]
    
    def initialize_options(self): 
        # pylint: disable=W0201
        self.bdist_base = 'build'
        self.dist_dir = 'dist'
        self.bdist_base_pa = os.path.join(self.bdist_base, 'TaskCoachPortable')
        self.name = self.version = self.license = self.url = self.filename = self.date = None
    
    def finalize_options(self):
        mandatoryOptions = [('name', 'the name of the application'),
                            ('version', 'the version of the application'),
                            ('license', 'the title of the license'),
                            ('url', 'the url of the application homepage'),
                            ('filename', 'the filename of the application without extension'),
                            ('date', 'the release date')]
        for option, description in mandatoryOptions:
            if not getattr(self, option):
                raise errors.DistutilsOptionError, \
                    'you must provide %s (--%s)'%(description, option)

    def run(self):
        self.create_portableapps_paths()
        self.copy_launcher()
        self.copy_appinfo()
        self.copy_defaultdata()
        self.copy_bin()
        self.copy_other()
        
    def create_portableapps_paths(self):
        for pathComponents in [('App', 'AppInfo'),
                               ('App', 'DefaultData', 'settings'),
                               ('App', 'TaskCoach'), 
                               ('Data', ),
                               ('Other', 'Help', 'images'), 
                               ('Other', 'Source')]:
            path = os.path.join(self.bdist_base_pa, *pathComponents)
            if not os.path.exists(path):
                os.makedirs(path)
                    
    def copy_launcher(self):
        launcher_src = os.path.join('build.in', 'portableapps')
        self.copy_files(launcher_src, self.bdist_base_pa)
               
    def copy_appinfo(self):
        src = os.path.join('build.in', 'portableapps', 'App', 'AppInfo')
        dest = os.path.join(self.bdist_base_pa, 'App', 'AppInfo')
        self.copy_files(src, dest)
        
    def copy_defaultdata(self):
        src = os.path.join('build.in', 'portableapps', 'App', 'DefaultData', 'settings')
        dest = os.path.join(self.bdist_base_pa, 'App', 'DefaultData', 'settings')
        self.copy_files(src, dest)

    def copy_bin(self):
        src = os.path.join(self.bdist_base, 'TaskCoach-%s-win32exe'%self.version)
        dest = os.path.join(self.bdist_base_pa, 'App', 'TaskCoach')
        self.copy_files(src, dest, copy_recursively=True)
        
    def copy_other(self):
        src = os.path.join('build.in', 'portableapps', 'Other')
        dest = os.path.join(self.bdist_base_pa, 'Other')
        self.copy_files(src, dest, copy_recursively=True)
            




'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


from taskcoachlib import meta
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty import chardet
import wx
import csv
import tempfile
import wx.grid as gridlib
import wx.wizard as wiz


class CSVDialect(csv.Dialect):
    def __init__(self, delimiter=',', quotechar='"', doublequote=True, escapechar=''):
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.quoting = csv.QUOTE_MINIMAL
        self.lineterminator = '\r\n'
        self.doublequote = doublequote
        self.escapechar = escapechar

        csv.Dialect.__init__(self)


class CSVImportOptionsPage(wiz.WizardPageSimple):
    def __init__(self, filename, *args, **kwargs):
        super(CSVImportOptionsPage, self).__init__(*args, **kwargs)

        self.delimiter = wx.Choice(self, wx.ID_ANY)
        self.delimiter.Append(_('Comma'))
        self.delimiter.Append(_('Tab'))
        self.delimiter.Append(_('Space'))
        self.delimiter.Append(_('Colon'))
        self.delimiter.Append(_('Semicolon'))
        self.delimiter.Append(_('Pipe'))
        self.delimiter.SetSelection(0)

        self.date = wx.Choice(self)
        self.date.Append(_('DD/MM (day first)'))
        self.date.Append(_('MM/DD (month first)'))
        self.date.SetSelection(0)
        
        self.quoteChar = wx.Choice(self, wx.ID_ANY)
        self.quoteChar.Append(_('Simple quote'))
        self.quoteChar.Append(_('Double quote'))
        self.quoteChar.SetSelection(1)

        self.quotePanel = wx.Panel(self, wx.ID_ANY)
        self.doubleQuote = wx.RadioButton(self.quotePanel, wx.ID_ANY, _('Double it'))
        self.doubleQuote.SetValue(True)
        self.escapeQuote = wx.RadioButton(self.quotePanel, wx.ID_ANY, _('Escape with'))
        self.escapeChar = wx.TextCtrl(self.quotePanel, wx.ID_ANY, '\\', size=(50, -1))
        self.escapeChar.Enable(False)
        self.escapeChar.SetMaxLength(1)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        hsizer.Add(self.doubleQuote, 1, wx.ALL, 3)
        hsizer.Add(self.escapeQuote, 1, wx.ALL, 3)
        hsizer.Add(self.escapeChar, 1, wx.ALL, 3)
        self.quotePanel.SetSizer(hsizer)
        
        self.importSelectedRowsOnly = wx.CheckBox(self, wx.ID_ANY, _('Import only the selected rows'))
        self.importSelectedRowsOnly.SetValue(False)

        self.hasHeaders = wx.CheckBox(self, wx.ID_ANY, _('First line describes fields'))
        self.hasHeaders.SetValue(True)

        self.grid = gridlib.Grid(self)
        self.grid.SetRowLabelSize(0)
        self.grid.SetColLabelSize(0)
        self.grid.CreateGrid(0, 0)
        self.grid.EnableEditing(False)
        self.grid.SetSelectionMode(self.grid.wxGridSelectRows)

        vsizer = wx.BoxSizer(wx.VERTICAL)
        gridSizer = wx.FlexGridSizer(0, 2)

        gridSizer.Add(wx.StaticText(self, wx.ID_ANY, _('Delimiter')), 0, 
                      wx.ALIGN_CENTRE_VERTICAL | wx.ALL, 3)
        gridSizer.Add(self.delimiter, 0, wx.ALL, 3)
        
        gridSizer.Add(wx.StaticText(self, wx.ID_ANY, _('Date format')), 0, 
                      wx.ALIGN_CENTER_VERTICAL | wx.ALL, 3)
        gridSizer.Add(self.date, 0, wx.ALL, 3)
        
        gridSizer.Add(wx.StaticText(self, wx.ID_ANY, _('Quote character')), 0, 
                      wx.ALIGN_CENTRE_VERTICAL | wx.ALL, 3)
        gridSizer.Add(self.quoteChar, 0, wx.ALL, 3)

        gridSizer.Add(wx.StaticText(self, wx.ID_ANY, _('Escape quote')), 0, 
                      wx.ALIGN_CENTRE_VERTICAL | wx.ALL, 3)
        gridSizer.Add(self.quotePanel, 0, wx.ALL, 3)

        gridSizer.Add(self.importSelectedRowsOnly, 0, wx.ALL, 3)
        gridSizer.Add((0, 0))

        gridSizer.Add(self.hasHeaders, 0, wx.ALL, 3)
        gridSizer.Add((0, 0))

        gridSizer.AddGrowableCol(1)
        vsizer.Add(gridSizer, 0, wx.EXPAND | wx.ALL, 3)

        vsizer.Add(self.grid, 1, wx.EXPAND | wx.ALL, 3)

        self.SetSizer(vsizer)

        self.headers = None

        self.filename = filename
        self.encoding = chardet.detect(file(filename, 'rb').read())['encoding']
        self.OnOptionChanged(None)

        wx.EVT_CHOICE(self.delimiter, wx.ID_ANY, self.OnOptionChanged)
        wx.EVT_CHOICE(self.quoteChar, wx.ID_ANY, self.OnOptionChanged)
        wx.EVT_CHECKBOX(self.importSelectedRowsOnly, wx.ID_ANY, self.OnOptionChanged)
        wx.EVT_CHECKBOX(self.hasHeaders, wx.ID_ANY, self.OnOptionChanged)
        wx.EVT_RADIOBUTTON(self.doubleQuote, wx.ID_ANY, self.OnOptionChanged)
        wx.EVT_RADIOBUTTON(self.escapeQuote, wx.ID_ANY, self.OnOptionChanged)
        wx.EVT_TEXT(self.escapeChar, wx.ID_ANY, self.OnOptionChanged)

    def OnOptionChanged(self, event):  # pylint: disable=W0613
        self.escapeChar.Enable(self.escapeQuote.GetValue())

        if self.filename is None:
            self.grid.SetRowLabelSize(0)
            self.grid.SetColLabelSize(0)
            if self.grid.GetNumberCols():
                self.grid.DeleteRows(0, self.grid.GetNumberRows())
                self.grid.DeleteCols(0, self.grid.GetNumberCols())
        else:
            if self.doubleQuote.GetValue():
                doublequote = True
                escapechar = ''
            else:
                doublequote = False
                escapechar = self.escapeChar.GetValue().encode('UTF-8')
            self.dialect = CSVDialect(delimiter={0: ',', 1: '\t', 2: ' ', 3: ':', 4: ';', 5: '|'}[self.delimiter.GetSelection()],
                                      quotechar={0: "'", 1: '"'}[self.quoteChar.GetSelection()],
                                      doublequote=doublequote, escapechar=escapechar)

            fp = tempfile.TemporaryFile()
            try:
                fp.write(file(self.filename, 'rU').read().decode(self.encoding).encode('UTF-8'))
                fp.seek(0)

                reader = csv.reader(fp, dialect=self.dialect)

                if self.hasHeaders.GetValue():
                    self.headers = [header.decode('UTF-8') for header in reader.next()]
                else:
                    # In some cases, empty fields are omitted if they're at the end...
                    hsize = 0
                    for line in reader:
                        hsize = max(hsize, len(line))
                    self.headers = [_('Field #%d') % idx for idx in xrange(hsize)]
                    fp.seek(0)
                    reader = csv.reader(fp, dialect=self.dialect)

                if self.grid.GetNumberCols():
                    self.grid.DeleteRows(0, self.grid.GetNumberRows())
                    self.grid.DeleteCols(0, self.grid.GetNumberCols())
                self.grid.InsertCols(0, len(self.headers))

                self.grid.SetColLabelSize(20)
                for idx, header in enumerate(self.headers):
                    self.grid.SetColLabelValue(idx, header)

                lineno = 0
                for line in reader:
                    self.grid.InsertRows(lineno, 1)
                    for idx, value in enumerate(line):
                        if idx < self.grid.GetNumberCols():
                            self.grid.SetCellValue(lineno, idx, value.decode('UTF-8'))
                    lineno += 1
            finally:
                fp.close()

    def GetOptions(self):
        return dict(dialect=self.dialect,
                    dayfirst=self.date.GetSelection() == 0,
                    importSelectedRowsOnly=self.importSelectedRowsOnly.GetValue(),
                    selectedRows=self.GetSelectedRows(),
                    hasHeaders=self.hasHeaders.GetValue(),
                    filename=self.filename,
                    encoding=self.encoding,
                    fields=self.headers)
        
    def GetSelectedRows(self):
        startRows = [row for row, dummy_column in self.grid.GetSelectionBlockTopLeft()]
        stopRows = [row for row, dummy_column in self.grid.GetSelectionBlockBottomRight()]
        selectedRows = []
        for startRow, stopRow in zip(startRows, stopRows):
            selectedRows.extend(range(startRow, stopRow + 1))
        return selectedRows

    def CanGoNext(self):
        if self.filename is not None:
            self.GetNext().SetOptions(self.GetOptions())
            return True, None
        return False, _('Please select a file.')


class CSVImportMappingPage(wiz.WizardPageSimple):
    def __init__(self, *args, **kwargs):
        super(CSVImportMappingPage, self).__init__(*args, **kwargs)

        # (field name, multiple values allowed)

        self.fields = [
            (_('None'), True),
            (_('ID'), False),
            (_('Subject'), False),
            (_('Description'), True),
            (_('Category'), True),
            (_('Priority'), False),
            (_('Planned start date'), False),
            (_('Due date'), False),
            (_('Actual start date'), False),
            (_('Completion date'), False),
            (_('Reminder date'), False),
            (_('Budget'), False),
            (_('Fixed fee'), False),
            (_('Hourly fee'), False),
            (_('Percent complete'), False),
            ]
        self.choices = []
        self.interior = wx.ScrolledWindow(self)
        self.interior.EnableScrolling(False, True)
        self.interior.SetScrollRate(10, 10)

        sizer = wx.BoxSizer()
        sizer.Add(self.interior, 1, wx.EXPAND)
        self.SetSizer(sizer)

    def SetOptions(self, options):
        self.options = options

        if self.interior.GetSizer():
            self.interior.GetSizer().Clear(True)

        for child in self.interior.GetChildren():
            self.interior.RemoveChild(child)
        self.choices = []

        gsz = wx.FlexGridSizer(0, 2, 4, 2)
        
        gsz.Add(wx.StaticText(self.interior, wx.ID_ANY, _('Column header in CSV file')))
        gsz.Add(wx.StaticText(self.interior, wx.ID_ANY, _('%s attribute') % meta.name))
        gsz.AddSpacer((3, 3))
        gsz.AddSpacer((3, 3))
        tcFieldNames = [field[0] for field in self.fields]
        for fieldName in options['fields']:
            gsz.Add(wx.StaticText(self.interior, wx.ID_ANY, fieldName), flag=wx.ALIGN_CENTER_VERTICAL)
            
            choice = wx.Choice(self.interior, wx.ID_ANY)
            for tcFieldName in tcFieldNames:
                choice.Append(tcFieldName)
            choice.SetSelection(self.findFieldName(fieldName, tcFieldNames))
            self.choices.append(choice)

            gsz.Add(choice, flag=wx.ALIGN_CENTER_VERTICAL)

        gsz.AddGrowableCol(1)
        self.interior.SetSizer(gsz)
        gsz.Layout()
        
    def findFieldName(self, fieldName, fieldNames):
        def fieldNameIndex(fieldName, fieldNames):
            return fieldNames.index(fieldName) if fieldName in fieldNames else 0
        
        index = fieldNameIndex(fieldName, fieldNames)
        return index if index else fieldNameIndex(fieldName[:6], [fieldName[:6] for fieldName in fieldNames])
        
    def CanGoNext(self):
        wrongFields = []
        countNotNone = 0

        for index, (fieldName, canMultiple) in enumerate(self.fields):
            count = 0
            for choice in self.choices:
                if choice.GetSelection() == index:
                    count += 1
                if choice.GetSelection() != 0:
                    countNotNone += 1
            if count > 1 and not canMultiple:
                wrongFields.append(fieldName)

        if countNotNone == 0:
            return False, _('No field mapping.')

        if len(wrongFields) == 1:
            return False, _('The "%s" field cannot be selected several times.') % wrongFields[0]

        if len(wrongFields):
            return False, _('The fields %s cannot be selected several times.') % ', '.join(['"%s"' % fieldName for fieldName in wrongFields])

        return True, None

    def GetOptions(self):
        options = dict(self.options)
        options['mappings'] = [self.fields[choice.GetSelection()][0] for choice in self.choices]
        return options


class CSVImportWizard(wiz.Wizard):
    def __init__(self, filename, *args, **kwargs):
        kwargs['style'] = wx.RESIZE_BORDER | wx.DEFAULT_DIALOG_STYLE
        super(CSVImportWizard, self).__init__(*args, **kwargs)

        self.optionsPage = CSVImportOptionsPage(filename, self)
        self.mappingPage = CSVImportMappingPage(self)
        self.optionsPage.SetNext(self.mappingPage)
        self.mappingPage.SetPrev(self.optionsPage)

        self.SetPageSize((600, -1))  # I know this is obsolete but it's the only one that works...

        wiz.EVT_WIZARD_PAGE_CHANGING(self, wx.ID_ANY, self.OnPageChanging)
        wiz.EVT_WIZARD_PAGE_CHANGED(self, wx.ID_ANY, self.OnPageChanged)

    def OnPageChanging(self, event):
        if event.GetDirection():
            can, msg = event.GetPage().CanGoNext()
            if not can:
                wx.MessageBox(msg, _('Information'), wx.OK)
                event.Veto()

    def OnPageChanged(self, event):
        if event.GetPage() == self.optionsPage:
            pass  # XXXTODO

    def RunWizard(self):
        return super(CSVImportWizard, self).RunWizard(self.optionsPage)

    def GetOptions(self):
        return self.mappingPage.GetOptions()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib.gui import dialog
from taskcoachlib import config
from wx.lib import sized_controls


class DummyColumn(object):
    def __init__(self, name):
        self.__name = name

    def name(self):
        return self.__name

    def header(self):
        return 'dummy column "%s"' % self.__name


class DummyViewer(object):
    def __init__(self):
        self.col1 = DummyColumn('one')
        self.col2 = DummyColumn('two')

    def title(self):
        return 'viewer'
    
    def hasHideableColumns(self):
        return True
    
    def visibleColumns(self):
        return []
    
    def columns(self):
        return [self.col1, self.col2]

    def selectableColumns(self):
        return [self.col2]


class DummyViewerContainer(object):
    def activeViewer(self):
        return 1
    
    def __getitem__(self, index):
        if index == 0:
            return DummyViewer()
        else:
            raise IndexError


class ExportDialogTest(test.wxTestCase):
    def testCreate(self):
        self.frame.viewer = DummyViewerContainer()
        settings = config.Settings(load=False)
        dialog.export.ExportAsHTMLDialog(self.frame, settings=settings)
        
        
class ColumnPickerTest(test.wxTestCase):
    def testCreate(self):
        panel = sized_controls.SizedPanel(self.frame)
        dialog.export.ColumnPicker(panel, DummyViewer())
        
    def testOnlySelectableColumns(self):
        panel = sized_controls.SizedPanel(self.frame)
        dlg = dialog.export.ColumnPicker(panel, DummyViewer())
        self.assertEqual(dlg.columnPicker.GetCount(), 1)
        self.assertEqual(dlg.columnPicker.GetClientData(0).name(), 'two')



'''
Task Coach - Your friendly task manager
Copyright (C) 2012 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.widgets import balloontip
import wx


class BalloonTipManager(balloontip.BalloonTipManager):
    def AddBalloonTip(self, settings, name, target, message=None, title=None, getRect=None):
        if settings.getboolean('balloontips', name):
            super(BalloonTipManager, self).AddBalloonTip(target, message=message, title=title,
                        bitmap=wx.ArtProvider.GetBitmap('lamp_icon', wx.ART_MENU, (16, 16)),
                        getRect=getRect, name=name, settings=settings)

    def OnBalloonTipShow(self, name=None, settings=None):
        settings.setboolean('balloontips', name, False)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system
from taskcoachlib.gui import menu
from taskcoachlib.thirdparty.pubsub import pub
import taskcoachlib.thirdparty.aui as aui
import wx


class ViewerContainer(object):
    ''' ViewerContainer is a container of viewers. It has a containerWidget
        that displays the viewers. The containerWidget is assumed to be 
        an AUI managed frame. The ViewerContainer knows which of its viewers
        is active and dispatches method calls to the active viewer or to the
        first viewer that can handle the method. This allows other GUI 
        components, e.g. menu's, to talk to the ViewerContainer as were
        it a regular viewer. '''
        
    def __init__(self, containerWidget, settings, *args, **kwargs):
        self.containerWidget = containerWidget
        self._notifyActiveViewer = False
        self.__bind_event_handlers()
        self._settings = settings
        self.viewers = []
        super(ViewerContainer, self).__init__(*args, **kwargs)

    def componentsCreated(self):
        self._notifyActiveViewer = True

    def advanceSelection(self, forward):
        ''' Activate the next viewer if forward is true else the previous 
            viewer. '''
        if len(self.viewers) <= 1:
            return  # Not enough viewers to advance selection
        active_viewer = self.activeViewer()
        current_index = self.viewers.index(active_viewer) if active_viewer else 0
        minimum_index, maximum_index = 0, len(self.viewers) - 1
        if forward:
            new_index = current_index + 1 if minimum_index <= current_index < maximum_index else minimum_index
        else:
            new_index = current_index - 1 if minimum_index < current_index <= maximum_index else maximum_index
        self.activateViewer(self.viewers[new_index])
        
    def isViewerContainer(self):
        ''' Return whether this is a viewer container or an actual viewer. '''
        return True

    def __bind_event_handlers(self):
        ''' Register for pane closing, activating and floating events. '''
        self.containerWidget.Bind(aui.EVT_AUI_PANE_CLOSE, self.onPageClosed)
        self.containerWidget.Bind(aui.EVT_AUI_PANE_ACTIVATED, 
                                  self.onPageChanged)
        self.containerWidget.Bind(aui.EVT_AUI_PANE_FLOATED, self.onPageFloated)
    
    def __getitem__(self, index):
        return self.viewers[index]
    
    def __len__(self):
        return len(self.viewers)

    def addViewer(self, viewer, floating=False):
        ''' Add a new pane with the specified viewer. '''
        self.containerWidget.addPane(viewer, viewer.title(), floating=floating)
        self.viewers.append(viewer)
        if len(self.viewers) == 1:
            self.activateViewer(viewer)
        pub.subscribe(self.onStatusChanged, viewer.viewerStatusEventType())
        
    def closeViewer(self, viewer):
        ''' Close the specified viewer. ''' 
        if viewer == self.activeViewer():
            self.advanceSelection(False)
        pane = self.containerWidget.manager.GetPane(viewer)
        self.containerWidget.manager.ClosePane(pane)
    
    def __getattr__(self, attribute):
        ''' Forward unknown attributes to the active viewer or the first
            viewer if there is no active viewer. '''
        return getattr(self.activeViewer() or self.viewers[0], attribute)

    def activeViewer(self):
        ''' Return the active (selected) viewer. '''
        all_panes = self.containerWidget.manager.GetAllPanes()
        for pane in all_panes:
            if pane.IsToolbar():
                continue
            if pane.HasFlag(pane.optionActive):
                if pane.IsNotebookControl():
                    notebook = aui.GetNotebookRoot(all_panes, pane.notebook_id)
                    return notebook.window.GetCurrentPage()
                else:
                    return pane.window
        return None
        
    def activateViewer(self, viewer_to_activate):
        ''' Activate (select) the specified viewer. '''
        self.containerWidget.manager.ActivatePane(viewer_to_activate)
        paneInfo = self.containerWidget.manager.GetPane(viewer_to_activate)
        if paneInfo.IsNotebookPage():
            self.containerWidget.manager.ShowPane(viewer_to_activate, True)
        self.sendViewerStatusEvent()

    def __del__(self):
        pass  # Don't forward del to one of the viewers.
    
    def onStatusChanged(self, viewer):
        if self.activeViewer() == viewer:
            self.sendViewerStatusEvent()
        pub.sendMessage('all.viewer.status', viewer=viewer)

    def onPageChanged(self, event):
        self.__ensure_active_viewer_has_focus()
        self.sendViewerStatusEvent()
        if self._notifyActiveViewer and self.activeViewer() is not None:
            self.activeViewer().activate()
        event.Skip()
    
    def sendViewerStatusEvent(self):
        pub.sendMessage('viewer.status')

    def __ensure_active_viewer_has_focus(self):
        if not self.activeViewer():
            return
        window = wx.Window.FindFocus()
        if operating_system.isMacOsXTiger_OrOlder() and window is None:
            # If the SearchCtrl has focus on Mac OS X Tiger,
            # wx.Window.FindFocus returns None. If we would continue,
            # the focus would be set to the active viewer right away,
            # making it impossible for the user to type in the search
            # control.
            return
        while window:
            if window == self.activeViewer():
                break
            window = window.GetParent()
        else:
            wx.CallAfter(self.activeViewer().SetFocus)

    def onPageClosed(self, event):
        if event.GetPane().IsToolbar():
            return
        window = event.GetPane().window
        if hasattr(window, 'GetPage'):
            # Window is a notebook, close each of its pages
            for pageIndex in range(window.GetPageCount()):
                self.__close_viewer(window.GetPage(pageIndex))
        else:
            # Window is a viewer, close it
            self.__close_viewer(window)
        # Make sure we have an active viewer
        if not self.activeViewer():
            self.activateViewer(self.viewers[0])
        event.Skip()
        
    def __close_viewer(self, viewer):
        ''' Close the specified viewer and unsubscribe all its event 
            handlers. '''
        # When closing an AUI managed frame, we get two close events, 
        # be prepared:
        if viewer in self.viewers:
            self.viewers.remove(viewer)
            viewer.detach()
        
    @staticmethod
    def onPageFloated(event):
        ''' Give floating pane accelerator keys for activating next and previous
            viewer. '''
        viewer = event.GetPane().window
        table = wx.AcceleratorTable([(wx.ACCEL_CTRL, wx.WXK_PAGEDOWN, 
                                      menu.activateNextViewerId),
                                     (wx.ACCEL_CTRL, wx.WXK_PAGEUP, 
                                      menu.activatePreviousViewerId)])
        viewer.SetAcceleratorTable(table)



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

# inotify wrapper from this recipe: http://code.activestate.com/recipes/576375-low-level-inotify-wrapper/
# Slightly modified to handle timeout and use select(); cleanup error handling

from twisted.internet.inotify import INotify
from twisted.python.filepath import FilePath

from taskcoachlib.filesystem import base
import os


class FilesystemNotifier(base.NotifierBase):
    def __init__(self):
        super(FilesystemNotifier, self).__init__()

        self.notifier = INotify()
        self.notifier.startReading()

    def setFilename(self, filename):
        if self._path is not None:
            self.notifier.ignore(FilePath(self._path))
        super(FilesystemNotifier, self).setFilename(filename)
        if self._path is not None:
            self.notifier.watch(FilePath(self._path), callbacks=[self.__notify])

    def stop(self):
        if self.notifier is not None:
            self.notifier.stopReading()
            self.notifier = None

    def __notify(self, handle, filepath, mask):
        myName = self._filename
        if myName is not None:
            if filepath.basename() == os.path.split(myName)[-1]:
                if self._check(myName) and myName:
                    self.stamp = os.stat(myName).st_mtime
                    self.onFileChanged()

    def onFileChanged(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from xml.etree import ElementTree as ET
from taskcoachlib import meta
from taskcoachlib.domain import date, task, note, category
from xml.etree import ElementTree as ET
import os
import sys


def flatten(elem):
    if len(elem) and not elem.text:
        elem.text = '\n'
    elif elem.text:
        elem.text = u'\n%s\n' % elem.text
    elem.tail = '\n'
    for child in elem:
        flatten(child)


class PIElementTree(ET.ElementTree):
    def __init__(self, pi, *args, **kwargs):
        self.__pi = pi
        ET.ElementTree.__init__(self, *args, **kwargs)

    def _write(self, file, node, encoding, namespaces):
        if node == self._root:
            # WTF? ElementTree does not write the encoding if it's ASCII or UTF-8...
            if encoding in ['us-ascii', 'utf-8']:
                file.write('<?xml version="1.0" encoding="%s"?>\n' % encoding)
            file.write(self.__pi.encode(encoding) + '\n')
        ET.ElementTree._write(self, file, node, encoding, namespaces)  # pylint: disable=E1101

    def write(self, file, encoding, *args, **kwargs):
        if encoding is None:
            encoding = 'utf-8'
        if sys.version_info >= (2, 7):
            file.write('<?xml version="1.0" encoding="%s"?>\n' % encoding)
            file.write(self.__pi.encode(encoding) + '\n')
            kwargs['xml_declaration'] = False
        ET.ElementTree.write(self, file, encoding, *args, **kwargs)


def sortedById(objects):
    s = [(obj.id(), obj) for obj in objects]
    s.sort()
    return [obj for dummy_id, obj in s]


class XMLWriter(object):
    maxDateTime = date.DateTime()
    
    def __init__(self, fd, versionnr=meta.data.tskversion):
        self.__fd = fd
        self.__versionnr = versionnr

    def write(self, taskList, categoryContainer,
              noteContainer, syncMLConfig, guid):
        root = ET.Element('tasks')

        for rootTask in sortedById(taskList.rootItems()):
            self.taskNode(root, rootTask)
        
        ownedNotes = self.notesOwnedByNoteOwners(taskList, categoryContainer)
        for rootCategory in sortedById(categoryContainer.rootItems()):
            self.categoryNode(root, rootCategory, taskList, noteContainer, ownedNotes)

        for rootNote in sortedById(noteContainer.rootItems()):
            self.noteNode(root, rootNote)
        
        if syncMLConfig:
            self.syncMLNode(root, syncMLConfig)
        if guid:
            ET.SubElement(root, 'guid').text = guid

        flatten(root)
        PIElementTree('<?taskcoach release="%s" tskversion="%d"?>\n' % (meta.data.version,
                                                                         self.__versionnr),
                                                root).write(self.__fd, 'utf-8')
    
    def notesOwnedByNoteOwners(self, *collectionOfNoteOwners):
        notes = []
        for noteOwners in collectionOfNoteOwners:
            for noteOwner in noteOwners:
                notes.extend(noteOwner.notes(recursive=True))
        return notes
    
    def taskNode(self, parentNode, task):  # pylint: disable=W0621
        maxDateTime = self.maxDateTime
        node = self.baseCompositeNode(parentNode, task, 'task', self.taskNode)
        node.attrib['status'] = str(task.getStatus())
        if task.plannedStartDateTime() != maxDateTime:
            node.attrib['plannedstartdate'] = str(task.plannedStartDateTime())
        if task.dueDateTime() != maxDateTime:
            node.attrib['duedate'] = str(task.dueDateTime())
        if task.actualStartDateTime() != maxDateTime:
            node.attrib['actualstartdate'] = str(task.actualStartDateTime())
        if task.completionDateTime() != maxDateTime:
            node.attrib['completiondate'] = str(task.completionDateTime())
        if task.percentageComplete():
            node.attrib['percentageComplete'] = str(task.percentageComplete())
        if task.recurrence():
            self.recurrenceNode(node, task.recurrence())
        if task.budget() != date.TimeDelta():
            node.attrib['budget'] = self.budgetAsAttribute(task.budget())
        if task.priority():
            node.attrib['priority'] = str(task.priority())
        if task.hourlyFee():
            node.attrib['hourlyFee'] = str(task.hourlyFee())
        if task.fixedFee():
            node.attrib['fixedFee'] = str(task.fixedFee())
        reminder = task.reminder() 
        if reminder != maxDateTime and reminder != None:
            node.attrib['reminder'] = str(reminder)
            reminderBeforeSnooze = task.reminder(includeSnooze=False)
            if reminderBeforeSnooze != None and reminderBeforeSnooze < task.reminder():
                node.attrib['reminderBeforeSnooze'] = str(reminderBeforeSnooze)
        prerequisiteIds = ' '.join([prerequisite.id() for prerequisite in \
            sortedById(task.prerequisites())])
        if prerequisiteIds:            
            node.attrib['prerequisites'] = prerequisiteIds
        if task.shouldMarkCompletedWhenAllChildrenCompleted() != None:
            node.attrib['shouldMarkCompletedWhenAllChildrenCompleted'] = \
                              str(task.shouldMarkCompletedWhenAllChildrenCompleted())
        for effort in sortedById(task.efforts()):
            self.effortNode(node, effort)
        for eachNote in sortedById(task.notes()):
            self.noteNode(node, eachNote)
        for attachment in sortedById(task.attachments()):
            self.attachmentNode(node, attachment)
        return node

    def recurrenceNode(self, parentNode, recurrence):
        attrs = dict(unit=recurrence.unit)
        if recurrence.amount > 1:
            attrs['amount'] = str(recurrence.amount)
        if recurrence.count > 0:
            attrs['count'] = str(recurrence.count)
        if recurrence.max > 0:
            attrs['max'] = str(recurrence.max)
        if recurrence.stop_datetime != self.maxDateTime:
            attrs['stop_datetime'] = str(recurrence.stop_datetime)
        if recurrence.sameWeekday:
            attrs['sameWeekday'] = 'True'
        if recurrence.recurBasedOnCompletion:
            attrs['recurBasedOnCompletion'] = 'True'
        return ET.SubElement(parentNode, 'recurrence', attrs)

    def effortNode(self, parentNode, effort):
        formattedStart = self.formatDateTime(effort.getStart())
        attrs = dict(id=effort.id(), status=str(effort.getStatus()), start=formattedStart)
        stop = effort.getStop()
        if stop != None:
            formattedStop = self.formatDateTime(stop)
            if formattedStop == formattedStart:
                # Make sure the effort duration is at least one second
                formattedStop = self.formatDateTime(stop + date.ONE_SECOND)
            attrs['stop'] = formattedStop
        node = ET.SubElement(parentNode, 'effort', attrs)
        if effort.description():
            ET.SubElement(node, 'description').text = effort.description()
        return node
    
    def categoryNode(self, parentNode, category, *categorizableContainers):  # pylint: disable=W0621
        def inCategorizableContainer(categorizable):
            for container in categorizableContainers:
                if categorizable in container:
                    return True
            return False
        node = self.baseCompositeNode(parentNode, category, 'category', self.categoryNode, 
                                      categorizableContainers)
        if category.isFiltered():
            node.attrib['filtered'] = str(category.isFiltered())
        if category.hasExclusiveSubcategories():
            node.attrib['exclusiveSubcategories'] = str(category.hasExclusiveSubcategories())
        for eachNote in sortedById(category.notes()):
            self.noteNode(node, eachNote)
        for attachment in sortedById(category.attachments()):
            self.attachmentNode(node, attachment)
        # Make sure the categorizables referenced are actually in the 
        # categorizableContainer, i.e. they are not deleted
        categorizableIds = ' '.join([categorizable.id() for categorizable in \
            sortedById(category.categorizables()) if inCategorizableContainer(categorizable)])
        if categorizableIds:            
            node.attrib['categorizables'] = categorizableIds
        return node
    
    def noteNode(self, parentNode, note):  # pylint: disable=W0621
        node = self.baseCompositeNode(parentNode, note, 'note', self.noteNode)
        for attachment in sortedById(note.attachments()):
            self.attachmentNode(node, attachment)
        return node

    def __baseNode(self, parentNode, item, nodeName):
        node = ET.SubElement(parentNode, nodeName,
                             dict(id=item.id(), status=str(item.getStatus())))
        if item.creationDateTime() > date.DateTime.min:
            node.attrib['creationDateTime'] = str(item.creationDateTime())
        if item.modificationDateTime() > date.DateTime.min:
            node.attrib['modificationDateTime'] = str(item.modificationDateTime())
        if item.subject():
            node.attrib['subject'] = item.subject()
        if item.description():
            ET.SubElement(node, 'description').text = item.description()
        return node

    def baseNode(self, parentNode, item, nodeName):
        ''' Create a node and add the attributes that all domain
            objects share, such as id, subject, description. '''
        node = self.__baseNode(parentNode, item, nodeName)
        if item.foregroundColor():
            node.attrib['fgColor'] = str(item.foregroundColor())
        if item.backgroundColor():
            node.attrib['bgColor'] = str(item.backgroundColor())
        if item.font():
            node.attrib['font'] = unicode(item.font().GetNativeFontInfoDesc())
        if item.icon():
            node.attrib['icon'] = str(item.icon())
        if item.selectedIcon():
            node.attrib['selectedIcon'] = str(item.selectedIcon())
        if item.ordering():
            node.attrib['ordering'] = str(item.ordering())
        return node

    def baseCompositeNode(self, parentNode, item, nodeName, childNodeFactory, childNodeFactoryArgs=()):
        ''' Same as baseNode, but also create child nodes by means of
            the childNodeFactory. '''
        node = self.__baseNode(parentNode, item, nodeName)
        if item.foregroundColor():
            node.attrib['fgColor'] = str(item.foregroundColor())
        if item.backgroundColor():
            node.attrib['bgColor'] = str(item.backgroundColor())
        if item.font():
            node.attrib['font'] = unicode(item.font().GetNativeFontInfoDesc())
        if item.icon():
            node.attrib['icon'] = str(item.icon())
        if item.selectedIcon():
            node.attrib['selectedIcon'] = str(item.selectedIcon())
        if item.ordering():
            node.attrib['ordering'] = str(item.ordering())
        if item.expandedContexts():
            node.attrib['expandedContexts'] = \
                     str(tuple(sorted(item.expandedContexts())))
        for child in sortedById(item.children()):
            childNodeFactory(node, child, *childNodeFactoryArgs)  # pylint: disable=W0142
        return node

    def attachmentNode(self, parentNode, attachment):
        node = self.baseNode(parentNode, attachment, 'attachment')
        node.attrib['type'] = attachment.type_
        data = attachment.data()
        if data is None:
            node.attrib['location'] = attachment.location()
        else:
            ET.SubElement(node, 'data', dict(extension=os.path.splitext(attachment.location())[-1])).text = \
                                data.encode('base64')
        for eachNote in sortedById(attachment.notes()):
            self.noteNode(node, eachNote)
        return node

    def syncMLNode(self, parentNode, syncMLConfig):
        node = ET.SubElement(parentNode, 'syncmlconfig')
        self.__syncMLNode(syncMLConfig, node)
        return node

    def __syncMLNode(self, cfg, node):
        for name, value in cfg.properties():
            ET.SubElement(node, 'property', dict(name=name)).text = value

        for childCfg in cfg.children():
            child = ET.SubElement(node, childCfg.name)
            self.__syncMLNode(childCfg, child)

    def budgetAsAttribute(self, budget):
        return '%d:%02d:%02d' % budget.hoursMinutesSeconds()

    def formatDateTime(self, dateTime):
        return dateTime.strftime('%Y-%m-%d %H:%M:%S')


class ChangesXMLWriter(object):
    def __init__(self, fd):
        self.__fd = fd

    def write(self, allChanges):
        root = ET.Element('changes')
        if allChanges:
            for devName, monitor in allChanges.items():
                devNode = ET.SubElement(root, 'device')
                devNode.attrib['guid'] = monitor.guid()
                for id_, changes in monitor.allChanges().items():
                    objNode = ET.SubElement(devNode, 'obj')
                    objNode.attrib['id'] = id_
                    if changes:
                        objNode.text = ','.join(list(changes))

        tree = ET.ElementTree(root)
        tree.write(self.__fd)


class TemplateXMLWriter(XMLWriter):
    def write(self, tsk):  # pylint: disable=W0221
        super(TemplateXMLWriter, self).write(task.TaskList([tsk]),
                   category.CategoryList(),
                   note.NoteContainer(),
                   None, None)

    def taskNode(self, parentNode, task):  # pylint: disable=W0621
        node = super(TemplateXMLWriter, self).taskNode(parentNode, task)

        for name, getter in [('plannedstartdate', 'plannedStartDateTime'),
                             ('duedate', 'dueDateTime'),
                             ('completiondate', 'completionDateTime'),
                             ('reminder', 'reminder')]:
            if hasattr(task, name + 'tmpl'):
                value = getattr(task, name + 'tmpl') or None
            else:
                dateTime = getattr(task, getter)()
                if dateTime not in (None, date.DateTime()):
                    delta = dateTime - date.Now()
                    minutes = delta.days * 24 * 60 + round(delta.seconds / 60.)
                    if minutes < 0:
                        value = '%d minutes ago' % -minutes
                    else:
                        value = '%d minutes from now' % minutes
                else:
                    value = None

            if value is None:
                if name in node.attrib:
                    del node.attrib[name]
            else:
                node.attrib[name + 'tmpl'] = value

        return node



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os, test

# Tests are run with ./tests as current dir, but setup.py expects the project
# root folder to be the current dir. Work around that by changing
# the current dir while importing setup.py:
cwd = os.path.realpath(os.path.curdir)
os.chdir('..')
import setup
os.chdir(cwd)


class LineEndingsTest(test.TestCase):
    def testNoDOSLineEndingsInPythonScripts(self):
        ''' On Linux, scripts won't work if they have DOS line endings. '''
        scripts = [os.path.join(test.projectRoot, script) \
                   for script in setup.setupOptions['scripts'] \
                   if script.endswith('.py')]
        for script in scripts:
            self.failIf('\r\n' in file(script, 'rb').read(), 
                        '%s contains DOS line endings'%script)



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>
Copyright (C) 2008 Thomas Sonne Olesen <tpo@sonnet.dk>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import command, widgets, domain, render
from taskcoachlib.domain import effort, date
from taskcoachlib.domain.base import filter  # pylint: disable=W0622
from taskcoachlib.gui import uicommand, menu, dialog
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.pubsub import pub
import base
import mixin
import refresher
import wx


class EffortViewer(base.ListViewer, 
                   mixin.FilterableViewerForCategorizablesMixin, 
                   mixin.SortableViewerForEffortMixin, 
                   mixin.SearchableViewerMixin, base.SortableViewerWithColumns): 
    defaultTitle = _('Effort')
    defaultBitmap = 'clock_icon'
    SorterClass = effort.EffortSorter
    
    def __init__(self, parent, taskFile, settings, *args, **kwargs):        
        kwargs.setdefault('settingsSection', 'effortviewer')
        self.__tasksToShowEffortFor = kwargs.pop('tasksToShowEffortFor', [])
        self.aggregation = 'details'  # Temporary value, will be properly set below
        self.__hiddenWeekdayColumns = []
        self.__hiddenTotalColumns = []
        self.__columnUICommands = None
        self.__domainObjectsToView = None
        super(EffortViewer, self).__init__(parent, taskFile, settings, *args, **kwargs)
        self.secondRefresher = refresher.SecondRefresher(self,
            effort.Effort.trackingChangedEventType())
        self.aggregation = settings.get(self.settingsSection(), 'aggregation')
        self.__initModeToolBarUICommands()
        self.registerObserver(self.onAttributeChanged_Deprecated,
                              eventType=effort.Effort.appearanceChangedEventType())
        pub.subscribe(self.onRoundingChanged, 
                      'settings.%s.round' % self.settingsSection())
        pub.subscribe(self.onRoundingChanged, 
                      'settings.%s.alwaysroundup' % self.settingsSection())
        pub.subscribe(self.on_aggregation_changed, 
                      'settings.%s.aggregation' % self.settingsSection())
        pub.subscribe(self.onHourDisplayChanged,
                      'settings.feature.decimaltime')

    def selectableColumns(self):
        columns = list()
        for column in self.columns():
            if column.name().startswith('total') and self.aggregation == 'details':
                continue
            if column.name() in ['monday', 'tuesday', 'wednesday', 
                 'thursday', 'friday', 'saturday', 'sunday'] and self.aggregation != 'week':
                continue
            columns.append(column)
        return columns

    def tasksToShowEffortFor(self):
        return self.__tasksToShowEffortFor

    def onRoundingChanged(self, value):  # pylint: disable=W0613
        self.__initRoundingToolBarUICommands()
        self.refresh()
        
    def onHourDisplayChanged(self, value):
        self.refresh()

    def __initModeToolBarUICommands(self):
        self.aggregationUICommand.setChoice(self.aggregation)
        self.__initRoundingToolBarUICommands()
        
    def __initRoundingToolBarUICommands(self):
        aggregated = self.isShowingAggregatedEffort()
        rounding = self.__round_precision() if aggregated else 0
        self.roundingUICommand.setChoice(rounding)
        self.roundingUICommand.enable(aggregated)
        self.alwaysRoundUpUICommand.setValue(self.__always_round_up())
        self.alwaysRoundUpUICommand.enable(aggregated and rounding != 0)
        
    def domainObjectsToView(self):
        if self.__domainObjectsToView is None:
            if self.__displayingNewTasks():
                tasks = self.tasksToShowEffortFor()
            else:
                tasks = domain.base.SelectedItemsFilter(self.taskFile.tasks(), 
                                                        selectedItems=self.tasksToShowEffortFor())
            self.__domainObjectsToView = tasks
        return self.__domainObjectsToView
    
    def __displayingNewTasks(self):
        return any([task not in self.taskFile.tasks() for task in self.tasksToShowEffortFor()])
    
    def detach(self):
        super(EffortViewer, self).detach()
        self.secondRefresher.removeInstance()
            
    def isShowingEffort(self):
        return True
    
    def curselectionIsInstanceOf(self, class_):
        return class_ == effort.Effort
    
    def on_aggregation_changed(self, value):
        self.__show_effort_aggregation(value)
    
    def __show_effort_aggregation(self, aggregation):
        ''' Change the aggregation mode. Can be one of 'details', 'day', 'week'
            and 'month'. '''
        assert aggregation in ('details', 'day', 'week', 'month')
        self.aggregation = aggregation
        self._refresh()

    def _refresh(self, clear=False):
        if clear:
            self.__domainObjectsToView = None
        self.setPresentation(self.createSorter(self.createFilter(\
                             self.domainObjectsToView())))
        self.secondRefresher.updatePresentation()
        self.registerPresentationObservers()
        # Invalidate the UICommands used for the column popup menu:
        self.__columnUICommands = None
        # Clear the selection to remove the cached selection
        self.clear_selection()
        # If the widget is auto-resizing columns, turn it off temporarily to 
        # make removing/adding columns faster
        autoResizing = self.widget.IsAutoResizing()
        if autoResizing:
            self.widget.ToggleAutoResizing(False)
        # Refresh first so that the list control doesn't think there are more
        # efforts than there really are when switching from aggregate mode to
        # detail mode.
        self.refresh()
        self._showWeekdayColumns(show=self.aggregation == 'week')
        self._showTotalColumns(show=self.aggregation != 'details')
        if autoResizing:
            self.widget.ToggleAutoResizing(True)
        self.__initRoundingToolBarUICommands()
        pub.sendMessage('effortviewer.aggregation')
            
    def isShowingAggregatedEffort(self):
        return self.aggregation != 'details'
    
    def createFilter(self, taskList):
        ''' Return a class that filters the original list. In this case we
            create an effort aggregator that aggregates the effort records in
            the taskList, either individually (i.e. no aggregation), per day,
            per week, or per month. '''
        aggregation = self.settings.get(self.settingsSection(), 'aggregation')
        deletedFilter = filter.DeletedFilter(taskList)
        categoryFilter = super(EffortViewer, self).createFilter(deletedFilter)
        searchFilter = filter.SearchFilter(self.createAggregator(categoryFilter,
                                                                 aggregation))
        return searchFilter
    
    def createAggregator(self, taskList, aggregation):
        ''' Return an instance of a class that aggregates the effort records 
            in the taskList, either:
            - individually (aggregation == 'details'), 
            - per day (aggregation == 'day'), 
            - per week ('week'), or 
            - per month ('month'). '''
        if aggregation == 'details':
            aggregator = effort.EffortList(taskList)
        else:
            aggregator = effort.EffortAggregator(taskList, aggregation=aggregation)
        return aggregator
            
    def createWidget(self):
        imageList = self.createImageList()  # Has side-effects
        self._columns = self._createColumns()  # pylint: disable=W0201
        itemPopupMenu = menu.EffortPopupMenu(self.parent, self.taskFile.tasks(),
            self.taskFile.efforts(), self.settings, self)
        columnPopupMenu = menu.EffortViewerColumnPopupMenu(self)
        self._popupMenus.extend([itemPopupMenu, columnPopupMenu])
        widget = widgets.VirtualListCtrl(self, self.columns(), self.onSelect,
            uicommand.Edit(viewer=self),
            itemPopupMenu, columnPopupMenu,
            resizeableColumn=1, **self.widgetCreationKeywordArguments())
        widget.AssignImageList(imageList, wx.IMAGE_LIST_SMALL)  # pylint: disable=E1101
        return widget
    
    def _createColumns(self):
        # pylint: disable=W0142
        kwargs = dict(resizeCallback=self.onResizeColumn)
        return [widgets.Column(name, columnHeader, eventType, 
                renderCallback=renderCallback,
                sortCallback=sortCallback,
                width=self.getColumnWidth(name), **kwargs) \
            for name, columnHeader, eventType, renderCallback, sortCallback in \
            ('period', _('Period'), effort.Effort.durationChangedEventType(), 
             self.__renderPeriod, 
             uicommand.ViewerSortByCommand(viewer=self, value='period')),
            ('task', _('Task'), effort.Effort.taskChangedEventType(), 
             lambda effort: effort.task().subject(recursive=True), None),
            ('description', _('Description'), 
             effort.Effort.descriptionChangedEventType(), 
             lambda effort: effort.description(), None)] + \
            [widgets.Column('categories', _('Categories'),
             width=self.getColumnWidth('categories'),
             renderCallback=self.renderCategories, **kwargs)] + \
            [widgets.Column(name, columnHeader, eventType, 
             width=self.getColumnWidth(name),
             renderCallback=renderCallback, alignment=wx.LIST_FORMAT_RIGHT,
             **kwargs) \
            for name, columnHeader, eventType, renderCallback in \
            ('timeSpent', _('Time spent'), 
             effort.Effort.durationChangedEventType(), self.__renderTimeSpent),
            ('totalTimeSpent', _('Total time spent'), 
             effort.Effort.durationChangedEventType(), 
             self.__renderTotalTimeSpent),
            ('revenue', _('Revenue'), effort.Effort.revenueChangedEventType(), 
             self.__renderRevenue),
            ('totalRevenue', _('Total revenue'), 
             effort.Effort.revenueChangedEventType(), self.__renderTotalRevenue)] + \
             [widgets.Column(name, columnHeader, eventType, 
              renderCallback=renderCallback, alignment=wx.LIST_FORMAT_RIGHT,
              width=self.getColumnWidth(name), **kwargs) \
             for name, columnHeader, eventType, renderCallback in \
                ('monday', _('Monday'), 
                 effort.Effort.durationChangedEventType(),  
                 lambda effort: self.__renderTimeSpentOnDay(effort, 0)),
                ('tuesday', _('Tuesday'), 
                 effort.Effort.durationChangedEventType(),
                 lambda effort: self.__renderTimeSpentOnDay(effort, 1)),
                ('wednesday', _('Wednesday'), 
                 effort.Effort.durationChangedEventType(),  
                 lambda effort: self.__renderTimeSpentOnDay(effort, 2)),
                ('thursday', _('Thursday'), 
                 effort.Effort.durationChangedEventType(),  
                 lambda effort: self.__renderTimeSpentOnDay(effort, 3)),
                ('friday', _('Friday'), 
                 effort.Effort.durationChangedEventType(),  
                 lambda effort: self.__renderTimeSpentOnDay(effort, 4)),
                ('saturday', _('Saturday'), 
                 effort.Effort.durationChangedEventType(),  
                 lambda effort: self.__renderTimeSpentOnDay(effort, 5)),
                ('sunday', _('Sunday'), 
                 effort.Effort.durationChangedEventType(),  
                 lambda effort: self.__renderTimeSpentOnDay(effort, 6))      
             ]

    def _showWeekdayColumns(self, show=True):
        if show:
            columnsToShow = self.__hiddenWeekdayColumns[:]
            self.__hiddenWeekdayColumns = []
        else:
            self.__hiddenWeekdayColumns = columnsToShow = \
                [column for column in self.visibleColumns() \
                 if column.name() in ['monday', 'tuesday', 'wednesday', 
                 'thursday', 'friday', 'saturday', 'sunday']]
        for column in columnsToShow:
            self.showColumn(column, show, refresh=False)

    def _showTotalColumns(self, show=True):
        if show:
            columnsToShow = self.__hiddenTotalColumns[:]
            self.__hiddenTotalColumns = []
        else:
            self.__hiddenTotalColumns = columnsToShow = \
                [column for column in self.visibleColumns() \
                 if column.name().startswith('total')]
        for column in columnsToShow:
            self.showColumn(column, show, refresh=False)
            
    def getColumnUICommands(self):
        # Create new UI commands every time since the UI commands depend on the
        # aggregation mode
        columnUICommands = \
            [uicommand.ToggleAutoColumnResizing(viewer=self,
                                                settings=self.settings),
             None,
             uicommand.ViewColumn(menuText=_('&Description'),
                                  helpText=_('Show/hide description column'),
                                  setting='description', viewer=self),
             uicommand.ViewColumn(menuText=_('&Categories'),
                                  helpText=_('Show/hide categories column'),
                                  setting='categories', viewer=self),
             uicommand.ViewColumn(menuText=_('&Time spent'),
                                  helpText=_('Show/hide time spent column'),
                                  setting='timeSpent', viewer=self),
             uicommand.ViewColumn(menuText=_('&Revenue'),
                                  helpText=_('Show/hide revenue column'),
                                  setting='revenue', viewer=self),]
        if self.aggregation != 'details':
            columnUICommands.insert(5,
                uicommand.ViewColumn(menuText=_('&Total time spent'),
                                     helpText=_('Show/hide total time spent column'),
                                     setting='totalTimeSpent', viewer=self))
            columnUICommands.insert(7,
                uicommand.ViewColumn(menuText=_('&Total revenue'),
                                     helpText=_('Show/hide total revenue column'),
                                     setting='totalRevenue', viewer=self))
        if self.aggregation == 'week':
            columnUICommands.append(\
                uicommand.ViewColumns(menuText=_('Effort per weekday'),
                    helpText=_('Show/hide time spent per weekday columns'),
                    setting=['monday', 'tuesday', 'wednesday', 'thursday', 
                             'friday', 'saturday', 'sunday'],
                    viewer=self))
        return columnUICommands
    
    def createCreationToolBarUICommands(self):
        return (uicommand.EffortNew(viewer=self, effortList=self.presentation(),
                                    taskList=self.taskFile.tasks(), 
                                    settings=self.settings),)
        
    def createActionToolBarUICommands(self):
        tasks = self.taskFile.tasks()
        return (uicommand.EffortStartForEffort(viewer=self, taskList=tasks),
                uicommand.EffortStop(viewer=self,
                                     effortList=self.taskFile.efforts(), 
                                     taskList=tasks))
                
    def createModeToolBarUICommands(self):
        # These are instance variables so that the choice can be changed 
        # programmatically
        # pylint: disable=W0201
        self.aggregationUICommand = \
            uicommand.EffortViewerAggregationChoice(viewer=self, 
                                                    settings=self.settings)
        self.roundingUICommand = uicommand.RoundingPrecision(viewer=self, 
                                                             settings=self.settings)
        self.alwaysRoundUpUICommand = uicommand.AlwaysRoundUp(viewer=self, 
                                                              settings=self.settings)
        return (self.aggregationUICommand, self.roundingUICommand, 
                self.alwaysRoundUpUICommand)

    def supportsRounding(self):
        return True
    
    def getRoundingUICommands(self):
        return [uicommand.AlwaysRoundUp(viewer=self, settings=self.settings), 
                None] + \
               [uicommand.RoundBy(menuText=menuText, value=value, viewer=self, 
                                  settings=self.settings) \
                for (menuText, value) in zip(uicommand.RoundingPrecision.choiceLabels, 
                                             uicommand.RoundingPrecision.choiceData)]
               
    def hasModes(self):
        return True
    
    def getModeUICommands(self):
        return [_('Effort aggregation'), None] + \
            [uicommand.EffortViewerAggregationOption(menuText=menuText, 
                                                     value=value,
                                                     viewer=self,
                                                     settings=self.settings)
             for (menuText, value) in zip(uicommand.EffortViewerAggregationChoice.choiceLabels,
                                          uicommand.EffortViewerAggregationChoice.choiceData)]

    def getItemImages(self, index, column=0):  # pylint: disable=W0613
        return {wx.TreeItemIcon_Normal: -1}
    
    def curselection(self):
        selection = super(EffortViewer, self).curselection()
        if self.aggregation != 'details':
            selection = [anEffort for compositeEffort in selection \
                                  for anEffort in compositeEffort]
        return selection
    
    def isselected(self, item):
        ''' When this viewer is in aggregation mode, L{curselection}
            returns the actual underlying L{Effort} objects instead of
            aggregates. This is a problem e.g. when exporting only a
            selection, since items we're iterating over (aggregates) are
            never in curselection(). This method is used instead. It just
            ignores the overridden version of curselection. '''

        return item in super(EffortViewer, self).curselection()

    def __sumTimeSpent(self, efforts):
        td = date.TimeDelta()
        for effort in efforts:
            td = td + effort.duration()

        sumTimeSpent = render.timeSpent(td, showSeconds=self.__show_seconds(), decimal=self.settings.getboolean('feature', 'decimaltime'))

        if sumTimeSpent == '':
            if self.settings.getboolean('feature', 'decimaltime'):
                sumTimeSpent = '0.0'
            elif self.__show_seconds():
                sumTimeSpent = '0:00:00'
            else:
                sumTimeSpent = '0:00'
        return sumTimeSpent

    def statusMessages(self):
        status1 = _('Effort: %d selected, %d visible, %d total. Time spent: %s selected, %s visible, %s total') % \
            (len(self.curselection()), len(self.presentation()), 
             len(self.taskFile.efforts()), self.__sumTimeSpent(self.curselection()),
             self.__sumTimeSpent(self.presentation()),  self.__sumTimeSpent(self.taskFile.efforts()))
        status2 = _('Status: %d tracking') % \
            self.presentation().nrBeingTracked()
        return status1, status2
   
    def newItemDialog(self, *args, **kwargs):
        selectedTasks = kwargs.get('selectedTasks', [])
        bitmap = kwargs.get('bitmap', 'new')
        if not selectedTasks:
            subjectDecoratedTaskList = [(task.subject(recursive=True), task) \
                                        for task in self.tasksToShowEffortFor()]
            subjectDecoratedTaskList.sort() # Sort by subject
            selectedTasks = [subjectDecoratedTaskList[0][1]]
        return super(EffortViewer, self).newItemDialog(selectedTasks, 
                                                       bitmap=bitmap)
        
    def itemEditorClass(self):
        return dialog.editor.EffortEditor
    
    def newItemCommandClass(self):
        return command.NewEffortCommand
    
    def newSubItemCommandClass(self):
        pass  # efforts are not composite.

    def deleteItemCommandClass(self):
        return command.DeleteEffortCommand
    
    # Rendering
    
    periodRenderers = dict( \
        details=lambda anEffort, humanReadable=True: render.dateTimePeriod(anEffort.getStart(), 
                                                       anEffort.getStop(), humanReadable=humanReadable),
        day=lambda anEffort, humanReadable=True: render.date(anEffort.getStart(),
                                                             humanReadable=humanReadable),
        week=lambda anEffort, humanReadable=True: render.weekNumber(anEffort.getStart()),
        month=lambda anEffort, humanReadable=True: render.month(anEffort.getStart()))

    def __renderPeriod(self, anEffort, humanReadable=True):
        ''' Return the period the effort belongs to. This depends on the
            current aggregation. If this period is the same as the previous
            period, an empty string is returned. '''
        return '' if self.__hasRepeatedPeriod(anEffort) else \
            self.periodRenderers[self.aggregation](anEffort, humanReadable=humanReadable)
                    
    def __hasRepeatedPeriod(self, anEffort):
        ''' Return whether the effort has the same period as the previous 
            effort record. '''
        index = self.presentation().index(anEffort)
        previousEffort = index > 0 and self.presentation()[index-1] or None
        if not previousEffort:
            return False
        if anEffort.getStart() != previousEffort.getStart():
            # Starts are not equal, so period cannot be repeated
            return False  
        if self.isShowingAggregatedEffort():
            # Starts and length of period are equal, so period is repeated
            return True  
        # If we get here, we are in details mode and the starts are equal 
        # Period can only be repeated when the stop times are also equal
        return anEffort.getStop() == previousEffort.getStop()
    
    def __renderTimeSpent(self, anEffort):
        ''' Return a rendered version of the effort duration. '''
        kwargs = dict()
        if isinstance(anEffort, effort.BaseCompositeEffort):
            kwargs['rounding'] = self.__round_precision()
            kwargs['roundUp'] = self.__always_round_up()
        duration = anEffort.duration(**kwargs)
        # Check for aggregation because we never round in details mode
        if self.isShowingAggregatedEffort():
            duration = self.__round_duration(duration)
            showSeconds = self.__show_seconds()
        else:
            showSeconds = True

        return render.timeSpent(duration, showSeconds=showSeconds, decimal=self.settings.getboolean('feature', 'decimaltime'))

    def __renderTotalTimeSpent(self, anEffort):
        ''' Return a rendered version of the effort total duration (of 
            composite efforts). '''
        # No need to check for aggregation because this method is only used
        # in aggregated mode
        total_duration = anEffort.duration(recursive=True,
               rounding=self.__round_precision(), roundUp=self.__always_round_up())
        return render.timeSpent(total_duration, 
                            showSeconds=self.__show_seconds(),
                            decimal=self.settings.getboolean('feature', 'decimaltime'))
    
    def __renderTimeSpentOnDay(self, anEffort, dayOffset):
        ''' Return a rendered version of the duration of the effort on a
            specific day. '''
        kwargs = dict()
        if isinstance(anEffort, effort.BaseCompositeEffort):
            kwargs['rounding'] = self.__round_precision()
            kwargs['roundUp'] = self.__always_round_up()
        duration = anEffort.durationDay(dayOffset, **kwargs) \
            if self.aggregation == 'week' else date.TimeDelta()
        return render.timeSpent(self.__round_duration(duration), 
                                showSeconds=self.__show_seconds(),
                                decimal=self.settings.getboolean('feature', 'decimaltime'))

    def getItemTooltipData(self, item):
        result = super(EffortViewer, self).getItemTooltipData(item)
        if isinstance(item, effort.CompositeEffort) and len(item):
            details = [_('Details:')]
            for theEffort in item:
                details.append(u'%s (%s)' % (render.dateTimePeriod(theEffort.getStart(), theEffort.getStop(), humanReadable=True), self.__renderTimeSpent(theEffort)))
            result.append((None, details))
        return result

    @staticmethod
    def __renderRevenue(anEffort):
        ''' Return the revenue of the effort as a monetary value. '''
        return render.monetaryAmount(anEffort.revenue())
    
    @staticmethod
    def __renderTotalRevenue(anEffort):
        ''' Return the total revenue of the effort as a monetary value. '''
        return render.monetaryAmount(anEffort.revenue(recursive=True))
    
    def __round_duration(self, duration):
        ''' Round a duration with the current precision and direction (i.e.
            always up or not). '''
        return duration.round(seconds=self.__round_precision(), 
                              alwaysUp=self.__always_round_up())
    
    def __show_seconds(self):
        ''' Return whether the viewer is showing seconds as part of 
            durations. '''
        return self.__round_precision() == 0
     
    def __round_precision(self):
        ''' Return with what precision the viewer is rounding durations. '''
        return self.settings.getint(self.settingsSection(), 'round')

    def __always_round_up(self):
        ''' Return whether durations are always rounded up or not. '''
        return self.settings.getboolean(self.settingsSection(), 'alwaysroundup')


class EffortViewerForSelectedTasks(EffortViewer):
    defaultTitle = _('Effort for selected task(s)')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'effortviewerforselectedtasks')
        self.__viewerContainer = kwargs.pop('viewerContainer')
        self.__currentTaskViewer = self.__viewerContainer.activeViewer() if self.__viewerContainer.activeViewer().isShowingTasks() else None
        pub.subscribe(self.onTaskSelectionChanged, 'all.viewer.status')
        super(EffortViewerForSelectedTasks, self).__init__(*args, **kwargs)

    def tasksToShowEffortFor(self):
        if self.__currentTaskViewer is not None:
            return domain.task.TaskList(self.__currentTaskViewer.curselection())
        return []

    def onTaskSelectionChanged(self, viewer):
        if viewer.isShowingTasks():
            self.__currentTaskViewer = viewer
            self._refresh(clear=True)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx, StringIO, os
import test
from taskcoachlib import persistence, gui, config, render
from taskcoachlib.domain import task, category, effort, date
    
    
class HTMLWriterTestCase(test.wxTestCase):
    treeMode = 'Subclass responsibility'
    filename = 'Subclass responsibility'
    
    def setUp(self):
        super(HTMLWriterTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.fd = StringIO.StringIO()
        self.writer = persistence.HTMLWriter(self.fd, self.filename)
        self.taskFile = persistence.TaskFile()
        self.task = task.Task('Task subject')
        self.taskFile.tasks().append(self.task)
        self.viewer = self.createViewer()
        
    def tearDown(self):
        super(HTMLWriterTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()
        cssFilename = self.filename + '.css'
        if os.path.exists(cssFilename):
            os.remove(cssFilename)
            
    def createViewer(self):
        raise NotImplementedError # pragma: no cover

    def __writeAndRead(self, selectionOnly):
        self.writer.write(self.viewer, self.settings, selectionOnly, True)
        return self.fd.getvalue()
    
    def expectInHTML(self, *htmlFragments, **kwargs):
        selectionOnly = kwargs.pop('selectionOnly', False)
        html = self.__writeAndRead(selectionOnly)
        for htmlFragment in htmlFragments:
            self.failUnless(htmlFragment in html, 
                            '%s not in %s'%(htmlFragment, html))
    
    def expectNotInHTML(self, *htmlFragments, **kwargs):
        selectionOnly = kwargs.pop('selectionOnly', False)
        html = self.__writeAndRead(selectionOnly)
        for htmlFragment in htmlFragments:
            self.failIf(htmlFragment in html, '%s in %s'%(htmlFragment, html))

    def selectItem(self, items):
        self.viewer.widget.select(items)


class CommonTestsMixin(object):
    def testHTML(self):
        self.expectInHTML('<html>\n', '</html>\n')
        
    def testHeader(self):
        self.expectInHTML('  <head>\n', '  </head>\n')
        
    def testStyle(self):
        self.expectInHTML('    <style type="text/css">\n', '    </style>\n')
        
    def testBody(self):
        self.expectInHTML('  <body>\n', '  </body>\n')


class TaskWriterTestCase(HTMLWriterTestCase):
    def createViewer(self):
        self.settings.set('taskviewer', 'treemode', self.treeMode)
        return gui.viewer.TaskViewer(self.frame, self.taskFile, self.settings)


class TaskTestsMixin(CommonTestsMixin):
    def testTaskSubject(self):
        self.expectInHTML('>Task subject<')
        
    def testWriteSelectionOnly(self):
        self.expectNotInHTML('>Task subject<', selectionOnly=True)
        
    def testWriteSelectionOnly_SelectedChild(self):
        child = task.Task('Child')
        self.task.addChild(child)
        self.taskFile.tasks().append(child)
        self.selectItem([child])
        self.expectInHTML('>Task subject<')

    def testColumnStyle(self):
        self.expectInHTML('      .subject {text-align: left}\n')
        
    def testSortIndicator(self):
        self.expectInHTML('id="sorted"')
        if not self.filename:
            self.expectInHTML('<u>')
            
    def testTaskStatusStyle(self):
        self.expectInHTML('      .completed {color: #00FF00}\n')
        
    def testTaskStatusStyleWhenForegroundColorChangedInSettings(self):
        self.settings.set('fgcolor', 'completedtasks', str(wx.RED))
        self.expectInHTML('      .completed {color: #FF0000}\n')
        
    def testOverdueTask(self):
        self.task.setDueDateTime(date.Yesterday())
        fragment = '<tr class="overdue">' if self.filename else '<font color="#FF0000">Task subject</font>'
        self.expectInHTML(fragment)

    def testCompletedTask(self):
        self.task.setCompletionDateTime()
        if self.filename:
            self.expectInHTML('<tr class="completed">')
        else:
            self.expectInHTML('<font color="#00FF00">Task subject</font>')

    def testTaskDueSoon(self):
        self.task.setDueDateTime(date.Now() + date.ONE_HOUR)
        fragment = '<tr class="duesoon">' if self.filename else '<font color="#FF8000">Task subject</font>' 
        self.expectInHTML(fragment)
        
    def testInactiveTask(self):
        self.task.setPlannedStartDateTime(date.Tomorrow())
        fragment = '<tr class="inactive">' if self.filename else '<font color="#C0C0C0">Task subject</font>'
        self.expectInHTML(fragment)
        
    def testLateTask(self):
        self.task.setPlannedStartDateTime(date.Yesterday())
        fragment = '<tr class="late">' if self.filename else '<font color="#A020F0">Task subject</font>'
        self.expectInHTML(fragment)

    def testTaskBackgroundColor(self):
        self.task.setActualStartDateTime(date.Now())
        self.task.setBackgroundColor(wx.RED)
        fragment = '<tr class="active" style="background: #FF0000">' if self.filename else '<tr bgcolor="#FF0000">'
        self.expectInHTML(fragment)
        
    def testTaskHasCategoryBackgroundColor(self):
        self.task.setActualStartDateTime(date.Now())
        cat = category.Category('cat', bgColor=wx.RED)
        self.task.addCategory(cat)
        fragment = '<tr class="active" style="background: #FF0000">' if self.filename else '<tr bgcolor="#FF0000">'
        self.expectInHTML(fragment)

    def testCategoryBackgroundColorAsTuple(self):
        self.task.setActualStartDateTime(date.Now())
        cat = category.Category('cat', bgColor=(255, 0, 0, 0))
        self.task.addCategory(cat)
        if self.filename:
            self.expectInHTML('<tr class="active" style="background: #FF0000">')
        else:
            self.expectInHTML('<tr bgcolor="#FF0000">')

    def testCSSLink(self):
        if self.filename:
            self.expectInHTML('<link href="filename.css" rel="stylesheet" type="text/css" media="screen">')
        else:
            self.expectNotInHTML('stylesheet')
            
    def testOSErrorWhileWritingCSS(self):
        def open(*args): # pylint: disable=W0613,W0622
            raise IOError
        self.writer._writeCSS(open=open) # pylint: disable=W0212
        

class TaskListTestsMixin(object):
    def testTaskDescription(self):
        self.task.setDescription('Task description')
        self.viewer.showColumnByName('description')
        self.expectInHTML('>Task description<')
    
    def testTaskDescriptionWithNewLine(self):
        self.task.setDescription('Line1\nLine2')
        self.viewer.showColumnByName('description')
        self.expectInHTML('>Line1<br>Line2<')
        
    def testCreationDateTime(self):
        self.viewer.showColumnByName('creationDateTime')
        self.expectInHTML(render.dateTime(self.task.creationDateTime(), 
                                          humanReadable=False))
        
    def testMissingCreationDateTime(self):
        self.viewer.showColumnByName('creationDateTime')
        self.taskFile.tasks().append(task.Task(creationDateTime=date.DateTime.min))
        self.taskFile.tasks().remove(self.task)
        self.expectNotInHTML('1/1/1')
        
    def testModificationDateTime(self):
        self.task.setModificationDateTime(date.DateTime(2012, 1, 1, 10, 0, 0))
        self.viewer.showColumnByName('modificationDateTime')
        self.expectInHTML(render.dateTime(self.task.modificationDateTime(),
                                          humanReadable=False))
        
    def testMissingModificationDateTime(self):
        self.viewer.showColumnByName('modificationDateTime')
        self.expectInHTML(render.dateTime(self.task.modificationDateTime(),
                                          humanReadable=False))
        self.expectNotInHTML('1/1/1')

        
class TaskListExportTest(TaskTestsMixin, TaskListTestsMixin, TaskWriterTestCase):
    treeMode = 'False'
    filename = 'filename'


class TaskListPrintTest(TaskTestsMixin, TaskListTestsMixin, TaskWriterTestCase):
    treeMode = 'False'
    filename = ''
                      

class TaskTreeExportTest(TaskTestsMixin, TaskWriterTestCase):
    treeMode = 'True'
    filename = 'filename'


class TaskTreePrintTest(TaskTestsMixin, TaskWriterTestCase):
    treeMode = 'True'
    filename = ''
    

class EffortWriterTestCase(CommonTestsMixin, HTMLWriterTestCase):
    filename = 'filename'
    
    def setUp(self):
        super(EffortWriterTestCase, self).setUp()
        now = date.DateTime.now()
        self.task.addEffort(effort.Effort(self.task, start=now,
                                          stop=now + date.ONE_SECOND))

    def createViewer(self):
        return gui.viewer.EffortViewer(self.frame, self.taskFile, self.settings)

    def testTaskSubject(self):
        self.expectInHTML('>Task subject<')
        
    def testEffortDuration(self):
        self.expectInHTML('>0:00:01<')
        
    def testColumnStyle(self):
        self.expectInHTML('      .task {text-align: left}\n')
        
        
class CategoryWriterTestsMixin(CommonTestsMixin):
    def testCategorySubject(self):
        self.expectInHTML('>Category<')
        
    def testCategoryBackgroundColor(self):
        self.category.setBackgroundColor(wx.RED)
        if self.filename:
            self.expectInHTML('<tr style="background: #FF0000">')
        else:
            self.expectInHTML('<tr bgcolor="#FF0000">')
        

class CategoryWriterTestCase(HTMLWriterTestCase):
    def setUp(self):
        super(CategoryWriterTestCase, self).setUp()
        self.category = category.Category('Category')
        self.taskFile.categories().append(self.category)

    def createViewer(self):
        return gui.viewer.CategoryViewer(self.frame, self.taskFile, 
                                         self.settings)

        
class CategoryWriterExportTest(CategoryWriterTestsMixin, CategoryWriterTestCase):
    filename = 'filename'
        

class CategoryWriterPrintTest(CategoryWriterTestsMixin, CategoryWriterTestCase):
    filename = ''
    



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.i18n import _
from taskcoachlib import patterns
import base


class ToggleCategoryCommand(base.BaseCommand):
    plural_name = _('Toggle category')
    singular_name = _('Toggle category of "%s"')
    
    def __init__(self, *args, **kwargs):
        self.category = kwargs.pop('category')
        super(ToggleCategoryCommand, self).__init__(*args, **kwargs) 
        # Keep track of previous category per categorizable in case of mutual 
        # exclusive categories:
        self.__previous_categories = dict()
        # When some items are in the category and some are not, only toggle
        # the items that are not in the category.
        items_not_in_category = [item for item in self.items \
                                 if self.category not in item.categories()]
        if 0 < len(items_not_in_category) < len(self.items):
            self.items = items_not_in_category
        
    def do_command(self):
        super(ToggleCategoryCommand, self).do_command()
        self.toggle_category()
        
    def undo_command(self):
        super(ToggleCategoryCommand, self).undo_command()
        self.toggle_category()
        
    redo_command = do_command
    
    @patterns.eventSource    
    def toggle_category(self, event=None):
        for categorizable in self.items:
            if self.category in categorizable.categories():
                self.unlink_category(self.category, categorizable, event)
                self.relink_previous_categories(categorizable, event)
            else:
                self.link_category(self.category, categorizable, event)
                self.unlink_previous_categories(categorizable, event)
        
    def unlink_previous_categories(self, categorizable, event):
        ''' Remove categorizable from any mutually exclusive categories it might
            belong to. '''
        if self.category.isMutualExclusive():
            parent = self.category.parent()
            if parent in categorizable.categories() and not parent.isMutualExclusive():
                self.unlink_previous_category(parent, categorizable, event)
            else:
                self.unlink_previous_mutual_exclusive_category(self.category.siblings(recursive=True), categorizable, event)
        if self.category.hasExclusiveSubcategories():
            self.unlink_previous_mutual_exclusive_category(self.category.children(), categorizable, event)
                
    def unlink_previous_mutual_exclusive_category(self, categories, categorizable, event):
        ''' Look for the category that categorizable belongs to and remove
            categorizable from it. '''
        for category in categories:
            if categorizable in category.categorizables():
                self.unlink_previous_category(category, categorizable, event)
                
    def unlink_previous_category(self, category, categorizable, event):
        ''' Remove categorizable from category, but remember the category so
            it can be restored later. '''
        self.unlink_category(category, categorizable, event)
        self.__previous_categories.setdefault(categorizable, []).append(category)
            
    def relink_previous_categories(self, categorizable, event):
        ''' Re-add categorizable to its previous categories. ''' 
        if categorizable in self.__previous_categories:
            for previous_category in self.__previous_categories[categorizable]:
                self.link_category(previous_category, categorizable, event)

    def link_category(self, category, categorizable, event):
        ''' Make categorizable belong to category. '''
        category.addCategorizable(categorizable, event=event)
        categorizable.addCategory(category, event=event)
        
    def unlink_category(self, category, categorizable, event):
        ''' Make categorizable no longer belong to category. '''
        category.removeCategorizable(categorizable, event=event)
        categorizable.removeCategory(category, event=event)



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib.changes import ChangeMonitor
from taskcoachlib.domain.categorizable import CategorizableCompositeObject
from taskcoachlib.domain.category import Category
from taskcoachlib.patterns import ObservableList

class MonitorCategorizableTest(test.TestCase):
    def setUp(self):
        self.monitor = ChangeMonitor()
        self.monitor.monitorClass(CategorizableCompositeObject)

        self.obj = CategorizableCompositeObject(subject=u'Object')
        self.list = ObservableList()
        self.monitor.monitorCollection(self.list)
        self.list.append(self.obj)

        self.cat1 = Category(subject=u'Cat #1')
        self.cat2 = Category(subject=u'Cat #2')
        self.catList = ObservableList()
        self.catList.append(self.cat1)
        self.catList.append(self.cat2)

    def tearDown(self):
        self.monitor.unmonitorClass(CategorizableCompositeObject)
        self.monitor.unmonitorCollection(self.list)

    def testAddCategory(self):
        self.monitor.resetAllChanges()
        self.obj.addCategory(self.cat1)
        self.assertEqual(self.monitor.getChanges(self.obj), set(['__add_category:%s' % self.cat1.id()]))

    def testRemoveCategory(self):
        self.obj.addCategory(self.cat1)
        self.monitor.resetAllChanges()
        self.obj.removeCategory(self.cat1)
        self.assertEqual(self.monitor.getChanges(self.obj), set(['__del_category:%s' % self.cat1.id()]))

    def testRemoveBadCategory(self):
        self.obj.addCategory(self.cat1)
        self.monitor.resetAllChanges()
        self.obj.removeCategory(self.cat2)
        self.assertEqual(self.monitor.getChanges(self.obj), set())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

"""

Classes for storing Funambol client configuration

"""

class SyncMLConfigNode(object):
    def __init__(self, name):
        super(SyncMLConfigNode, self).__init__()

        self.name = name

        self.__children = []
        self.__properties = {}

    def children(self):
        return self.__children

    def properties(self):
        return self.__properties.items()

    def addChild(self, child):
        self.__children.append(child)

    def get(self, name):
        return self.__properties.get(name, '')

    def set(self, name, value):
        self.__properties[name] = value

    def __getitem__(self, name):
        for child in self.__children:
            if child.name == name:
                return child
        raise KeyError, name


def createDefaultSyncConfig(uid):
    cfg = SyncMLConfigNode('root')
    root = SyncMLConfigNode('TaskCoach-%s' % uid)
    cfg.addChild(root)
    spds = SyncMLConfigNode('spds')
    root.addChild(spds)
    sources = SyncMLConfigNode('sources')
    spds.addChild(sources)
    syncml = SyncMLConfigNode('syncml')
    spds.addChild(syncml)
    tasks = SyncMLConfigNode('TaskCoach-%s.Tasks' % uid)
    sources.addChild(tasks)
    notes = SyncMLConfigNode('TaskCoach-%s.Notes' % uid)
    sources.addChild(notes)
    auth = SyncMLConfigNode('Auth')
    syncml.addChild(auth)
    conn = SyncMLConfigNode('Conn')
    syncml.addChild(conn)

    return cfg



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from xml.etree import ElementTree as ET
import os


def anonymize(filename):
    ''' Anonymize the file specified by the filename by reading its contents,
        replacing the contents with X's and saving the anonymized contents to 
        a copy of the file. '''
    
    def anonymize_string(string):
        ''' Return an anonymized version of the string. '''
        return u'X' * len(string)
    
    def anonymize_text(text):
        ''' Return an anonymized version of the text, keeping the line 
            breaks. '''
        return '\n'.join([anonymize_string(line) for line in text.split('\n')])

    def anonymize_node(node):
        ''' Recursively anonymize the node and all of its child nodes. '''
        for child in node:
            anonymize_node(child)

        if 'subject' in node.attrib:
            node.attrib['subject'] = anonymize_string(node.attrib['subject'])

        if node.tag in ('description', 'data') and node.text:
            node.text = anonymize_text(node.text)
            if node.tag == 'data':
                node.attrib['extension'] = \
                    anonymize_string(node.attrib['extension'])

        if node.tag == 'property' and 'name' in node.attrib and \
           node.attrib['name'] == 'username':
            node.text = 'XXX'  # pylint: disable=W0511

        if node.tag == 'attachment' and 'location' in node.attrib:
            node.attrib['location'] = anonymize_string(node.attrib['location'])
            
    tree = ET.parse(file(filename, 'rb'))
    anonymize_node(tree.getroot())
    name, ext = os.path.splitext(filename)
    anonymized_filename = name + '.anonymized' + ext
    tree.write(anonymized_filename)
    return anonymized_filename



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import test
from taskcoachlib import gui, config, operating_system


class WindowDimensionsTrackerTest(test.wxTestCase):
    def setUp(self):
        super(WindowDimensionsTrackerTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.section = 'window'
        self.settings.setvalue(self.section, 'position', (50, 50))
        self.settings.setvalue(self.section, 'starticonized', 'Never')
        if operating_system.isWindows():
            self.frame.Show()
        self.tracker = gui.windowdimensionstracker.WindowDimensionsTracker( \
                           self.frame, self.settings)
        
    def test_initial_position(self):
        self.assertEqual(self.settings.getvalue(self.section, 'position'), 
                         self.frame.GetPositionTuple())
    
    def test_initial_size(self):
        # See MainWindowTest...
        width, height = self.frame.GetSizeTuple()
        if operating_system.isMac():  # pragma: no cover
            width, height = self.frame.GetClientSize()
            height -= 18 
        self.assertEqual((width, height), 
                         self.settings.getvalue(self.section, 'size'))
     
    @test.skipOnPlatform('__WXGTK__')
    def test_maximize(self):
        for maximized in [True, False]:
            self.frame.Maximize(maximized)
            self.assertEqual(maximized, self.frame.IsMaximized())
            self.assertEqual(maximized, 
                             self.settings.getboolean(self.section, 
                                                      'maximized'))
            
    def test_change_size(self):
        self.frame.Maximize(False)
        if operating_system.isMac():
            self.frame.SetClientSize((123, 200))
        else:
            self.frame.ProcessEvent(wx.SizeEvent((123, 200)))
        self.assertEqual((123, 200), 
                         self.settings.getvalue(self.section, 'size'))
        
    def test_move(self):
        self.frame.Maximize(False)
        self.frame.Iconize(False)
        self.frame.ProcessEvent(wx.MoveEvent((200, 200)))
        self.assertEqual((200, 200), 
                         self.settings.getvalue(self.section, 'position'))



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.powermgt.base import PowerStateMixinBase
# pylint: disable=F0401
import win32api
import win32gui
import win32con
import wx


class PowerStateMixin(PowerStateMixinBase):
    def __init__(self, *args, **kwargs):
        super(PowerStateMixin, self).__init__(*args, **kwargs)

        self.__oldProc = win32gui.SetWindowLong(self.GetHandle(),
                                                win32con.GWL_WNDPROC,
                                                self.__WndProc)

    def __WndProc(self, hWnd, msg, wParam, lParam):
        if msg == win32con.WM_DESTROY:
            win32api.SetWindowLong(self.GetHandle(),
                                   win32con.GWL_WNDPROC,
                                   self.__oldProc)

        if msg == win32con.WM_POWERBROADCAST:
            if wParam == win32con.PBT_APMSUSPEND:
                wx.CallAfter(self.OnPowerState, self.POWEROFF)
            elif wParam == win32con.PBT_APMRESUMESUSPEND:
                wx.CallAfter(self.OnPowerState, self.POWERON)

        return win32gui.CallWindowProc(self.__oldProc,
                                       hWnd, msg, wParam, lParam)
